"""
Absufyu: Converter
------------------
Convert stuff

Version: 5.17.0
Date updated: 20/11/2025 (dd/mm/yyyy)

Feature:
--------
- Text2Chemistry
- Str2Pixel
- Base64EncodeDecode
"""

# Module level
# ---------------------------------------------------------------------------
__all__ = [
    "Base64EncodeDecode",
    "Text2Chemistry",
    "Str2Pixel",
    # Support
    "ChemistryElement",
]


# Library
# ---------------------------------------------------------------------------
import base64
import math
import re
import string
from itertools import chain, combinations
from pathlib import Path
from typing import Self

from absfuyu.core.baseclass import BaseClass, CLITextColor
from absfuyu.core.docstring import versionadded
from absfuyu.logger import logger
from absfuyu.pkg_data import DataList, DataLoader
from absfuyu.util.text_table import OneColumnTableMaker


# Class
# ---------------------------------------------------------------------------
@versionadded("3.0.0")
class Base64EncodeDecode(BaseClass):
    """
    Encode and decode base64
    """

    @staticmethod
    def encode(data: str) -> str:
        """Base64 encode"""
        return base64.b64encode(data.encode()).decode()

    @staticmethod
    def decode(data: str) -> str:
        """Base64 decode"""
        return base64.b64decode(data).decode()

    @staticmethod
    @versionadded("4.1.0")
    def encode_image(img_path: Path | str, data_tag: bool = False) -> str:
        """
        Encode image file into base64 string

        Parameters
        ----------
        img_path : Path | str
            Path to image

        data_tag : bool, optional
            Add data tag before base64 string, by default ``False``

        Returns
        -------
        str
            Encoded image
        """
        img = Path(img_path)
        with open(img, "rb") as img_file:
            b64_data = base64.b64encode(img_file.read()).decode("utf-8")
        if data_tag:
            return f"data:image/{img.suffix[1:]};charset=utf-8;base64,{b64_data}"
        return b64_data


class ChemistryElement(BaseClass):
    """
    Chemistry Element

    Parameters
    ----------
    name : str
        Element name

    number : int
        Order in periodic table

    symbol : str
        Short symbol of element

    atomic_mass : float
        Atomic mass of element
    """

    __slots__ = ("name", "number", "symbol", "atomic_mass")

    def __init__(self, name: str, number: int, symbol: str, atomic_mass: float) -> None:
        """
        Chemistry Element

        Parameters
        ----------
        name : str
            Element name

        number : int
            Order in periodic table

        symbol : str
            Short symbol of element

        atomic_mass : float
            Atomic mass of element
        """
        self.name = name
        self.number = number
        self.symbol = symbol
        self.atomic_mass = atomic_mass

    def __str__(self) -> str:
        return f"{self.__class__.__name__}({self.symbol})"

    def to_dict(self) -> dict[str, str | int | float]:
        """
        Output content to dict

        Returns
        -------
        dict[str, str | int | float]
            Dict version of element
        """
        # return {"name": self.name, "number": self.number, "symbol": self.symbol, "atomic_mass": self.atomic_mass}
        return {x: getattr(self, x) for x in self.__slots__}

    @classmethod
    def from_dict(cls, data: dict[str, str | int | float]) -> Self:
        """
        Convert from ``dict`` data

        Parameters
        ----------
        data : dict[str, str | int | float]
            Dict data

        Returns
        -------
        Self
            ChemistryElement
        """
        return cls(
            name=data["name"],  # type: ignore
            number=int(data["number"]),
            symbol=data["symbol"],  # type: ignore
            atomic_mass=float(data["atomic_mass"]),
        )


class Text2Chemistry(BaseClass):
    """
    Convert text into chemistry symbols.
    """

    def __init__(self) -> None:
        self.data_location = DataList.CHEMISTRY

    def __str__(self) -> str:
        return f"{self.__class__.__name__}()"

    def _load_chemistry_data(self) -> list[ChemistryElement]:
        """
        Load chemistry pickle data
        """
        data: list[dict] = DataLoader(self.data_location).load()
        return [ChemistryElement.from_dict(x) for x in data]

    @property
    def unvailable_characters(self) -> set[str]:
        """
        Characters that can not be converted (unvailable chemistry symbol)

        Returns
        -------
        set[str]
            Set of unvailable characters
        """
        base = set(string.ascii_lowercase)
        available = set(
            "".join(map(lambda x: x.symbol.lower(), self._load_chemistry_data()))
        )
        # logger.debug(base)
        # logger.debug(available)
        return base.difference(available)

    def convert(self, text: str) -> list[list[ChemistryElement]] | list:
        """
        Convert text to chemistry symbol

        Parameters
        ----------
        text : str
            Desired text

        Returns
        -------
        list[list[ChemistryElement]] | list
            Converted text (empty list when failed to convert)

        Raises
        ------
        ValueError
            When text contains digit, whitespaces, ...
        """
        # Check if `text` is a word (without digits)
        is_word_pattern = r"^[a-zA-Z]+$"
        if re.search(is_word_pattern, text) is None:
            logger.error("Convert Failed. Word Only!")
            raise ValueError("Convert Failed. Word Only!")
        for x in self.unvailable_characters:
            if text.find(x) != -1:
                logger.debug(
                    f"{text} contains unvailable characters: {self.unvailable_characters}"
                )
                # raise ValueError(f"Text contains {self.unvailable_character}")
                return []

        # Setup
        text_lower = text.lower()
        data = self._load_chemistry_data()

        # List possible elements
        possible_elements: list[ChemistryElement] = []
        for i, letter in enumerate(text_lower):
            for element in data:
                if element.symbol.lower().startswith(
                    letter
                ):  # Check for `element.symbol` starts with `letter`
                    # logger.debug(f"{letter} {element}")
                    if element.symbol.lower().startswith(
                        text_lower[i : i + len(element.symbol)]
                    ):  # Check for `element.symbol` with len > 1 starts with `letter` of len(element.symbol)
                        possible_elements.append(element)
                    # Break when reach last letter in text
                    if letter == text_lower[-1]:
                        break
        logger.debug(possible_elements)
        if len(possible_elements) < 1:  # No possible elements
            return []

        # temp = []
        # for i in range(min_combination_range, len(text_lower)+1):
        #     comb = combinations(possible_elements, i)
        #     temp.append(comb)
        # possible_combinations = chain(*temp)
        max_symbol_len = max(
            map(lambda x: len(x.symbol), possible_elements)
        )  # Max len of `element.symbol`
        min_combination_range = math.ceil(len(text_lower) / max_symbol_len)
        logger.debug(f"Combination range: [{min_combination_range}, {len(text_lower)}]")
        possible_combinations = chain(
            *(
                combinations(possible_elements, i)
                for i in range(min_combination_range, len(text_lower) + 1)
            )
        )
        # logger.debug(list(possible_combinations))

        output = []
        for comb in possible_combinations:
            merged = "".join(map(lambda x: x.symbol, comb))
            if text_lower == merged.lower():
                output.append(list(comb))
                logger.debug(f"Found: {merged}")

        return output

    @staticmethod
    @versionadded("4.2.0")
    def beautify_result(
        result: list[list[ChemistryElement]] | list,
    ) -> str:
        """
        Beautify the result from ``Text2Chemistry.convert()``

        Parameters
        ----------
        result : list[list[ChemistryElement]] | list
            Convert ``Text2Chemistry.convert()`` result

        Returns
        -------
        str
            Beautified output
        """
        if len(result) == 0:
            res = "No possible combination"
        else:
            header = ["Text to Chemistry"]
            body = []
            max_table_len = len(header[0])

            for i, solution in enumerate(result, start=1):
                msg = []
                max_word_len = max([len(x.name) for x in solution])
                msg.append(f"Option {i:02}: {', '.join([x.symbol for x in solution])}")
                for x in solution:
                    msg.append(
                        f"{x.symbol.ljust(2)} "
                        f"({x.number:02}. {x.name.ljust(max_word_len)}"
                        f" - {round(x.atomic_mass, 2)})"
                    )
                body.append(msg)

            max_table_len = max(
                max([max([len(x) for x in opt]) for opt in body]), max_table_len
            )
            table = OneColumnTableMaker(ncols=max_table_len + 4, style="normal")
            table.add_paragraph(header)
            for x in body:
                table.add_paragraph(x)

            res = table.make_table()
        return res


class Str2Pixel(BaseClass):
    """
    Convert str into pixel

    Parameters
    ----------
    str_data : str
        | Pixel string data (Format: ``<number_of_pixel><color_code>``)
        | Example: ``50w20b`` = 50 white pixels and 20 black pixels

    pixel_size : int, optional
        Pixel size, by default ``2``

    pixel_symbol_overwrite : str | None, optional
        Overwrite pixel symbol, by default ``None``
    """

    PIXEL = "\u2588"

    def __init__(
        self,
        str_data: str,
        *,
        pixel_size: int = 2,
        pixel_symbol_overwrite: str | None = None,
    ) -> None:
        """
        Convert str into pixel

        Parameters
        ----------
        str_data : str
            | Pixel string data (Format: ``<number_of_pixel><color_code>``)
            | Example: ``50w20b`` = 50 white pixels and 20 black pixels

        pixel_size : int, optional
            Pixel size, by default ``2``

        pixel_symbol_overwrite : str | None, optional
            Overwrite pixel symbol, by default ``None``
        """
        self.data = str_data
        if pixel_symbol_overwrite is None:
            self.pixel = self.PIXEL * int(max(pixel_size, 1))
        else:
            self.pixel = pixel_symbol_overwrite

    def _extract_pixel(self):
        """Split str_data into corresponding int and str"""
        num = re.split("[a-zA-Z]", self.data)  # type: ignore
        num = filter(lambda x: x != "", num)  # type: ignore # Clean "" in list
        num = list(map(int, num))  # type: ignore
        char = re.split("[0-9]", self.data)
        char = filter(lambda x: x != "", char)  # type: ignore
        return [x for y in zip(num, char) for x in y]

    def convert(self, line_break: bool = True) -> str:
        """
        Convert data into pixel

        Parameters
        ----------
        line_break : bool, optional
            Add ``\\n`` at the end of line, by default ``True``

        Returns
        -------
        str
            Converted colored pixels
        """
        # Extract pixel
        pixel_map = self._extract_pixel()

        # Translation to color
        translate = {
            "w": CLITextColor.WHITE,
            "b": CLITextColor.BLACK,
            "B": CLITextColor.BLUE,
            "g": CLITextColor.GRAY,
            "G": CLITextColor.GREEN,
            "r": CLITextColor.RED,
            "R": CLITextColor.DARK_RED,
            "m": CLITextColor.MAGENTA,
            "y": CLITextColor.YELLOW,
            "E": CLITextColor.RESET,
            "N": "\n",  # New line
        }

        # import colorama
        # translate = {
        #     "w": colorama.Fore.WHITE,
        #     "b": colorama.Fore.BLACK,
        #     "B": colorama.Fore.BLUE,
        #     "g": colorama.Fore.LIGHTBLACK_EX, # Gray
        #     "G": colorama.Fore.GREEN,
        #     "r": colorama.Fore.LIGHTRED_EX,
        #     "R": colorama.Fore.RED, # Dark red
        #     "m": colorama.Fore.MAGENTA,
        #     "y": colorama.Fore.YELLOW,
        #     "E": colorama.Fore.RESET,
        #     "N": "\n", # New line
        # }

        # Output
        out = ""
        for i, x in enumerate(pixel_map):
            if isinstance(x, str):
                temp = self.pixel * pixel_map[i - 1]
                out += f"{translate[x]}{temp}{translate['E']}"
        if line_break:
            return out + "\n"
        else:
            return out



"""
Test: Tools

Version: 5.17.0
Date updated: 20/11/2025 (dd/mm/yyyy)
"""

import pytest

from absfuyu.tools.checksum import Checksum
from absfuyu.tools.shutdownizer import ShutDownizer


# MARK: checksum
@pytest.mark.abs_tools
class TestChecksum:
    """absfuyu.tools.checksum.Checksum"""

    def test_checksum(self) -> None:
        pass


# MARK: shutdownizer
@pytest.mark.abs_tools
class TestShutDownizer:
    """absfuyu.tools.shutdownizer.ShutDownizer"""

    def test_shutdown(self) -> None:
        ins = ShutDownizer()
        assert ins



"""
Test: Data extension - ListExt

Version: 5.17.0
Date updated: 20/11/2025 (dd/mm/yyyy)
"""

import pytest

from absfuyu.dxt import ListExt


@pytest.fixture
def list_example():
    return ListExt([3, 8, 5, "Test", "String", "ABC", [1, 2, 3], [0, 8, 6]])


@pytest.fixture
def list_example_2():
    return ListExt(["Test", "String", "ABC", "Tension", "Tent", "Strong"])


@pytest.mark.abs_dxt
class TestListExt:
    """absfuyu.dxt.ListExt"""

    # stringify
    def test_stringify(self, list_example: ListExt) -> None:
        assert all([isinstance(x, str) for x in list_example.stringify()]) is True

    # sorts
    def test_sorts(self, list_example: ListExt) -> None:
        assert list_example.sorts() == [
            3,
            5,
            8,
            "ABC",
            "String",
            "Test",
            [0, 8, 6],
            [1, 2, 3],
        ]

    # freq
    def test_freq(self, list_example_2: ListExt) -> None:
        assert list_example_2.freq(sort=True) == {
            "ABC": 1,
            "String": 1,
            "Strong": 1,
            "Tension": 1,
            "Tent": 1,
            "Test": 1,
        }

    def test_freq_2(self, list_example_2: ListExt) -> None:
        assert list_example_2.freq(sort=True, num_of_first_char=2) == {
            "AB": 1,
            "St": 2,
            "Te": 3,
        }

    def test_freq_3(self, list_example_2: ListExt) -> None:
        assert list_example_2.freq(
            sort=True, num_of_first_char=2, appear_increment=True
        ) == [1, 3, 6]

    # slice_points
    def test_slice_points(self, list_example_2: ListExt) -> None:
        assert list_example_2.slice_points([1, 3]) == [
            ["Test"],
            ["String", "ABC"],
            ["Tension", "Tent", "Strong"],
        ]

    # pick_one
    def test_pick_one(self, list_example_2: ListExt) -> None:
        assert len([list_example_2.pick_one()]) == 1

    def test_pick_one_error(self) -> None:
        """Empty list"""
        with pytest.raises(IndexError) as excinfo:
            ListExt([]).pick_one()
            assert str(excinfo.value)

    # len_items
    def test_len_items(self, list_example_2: ListExt) -> None:
        assert list_example_2.len_items() == [4, 6, 3, 7, 4, 6]

    # mean_len
    def test_mean_len(self, list_example_2: ListExt) -> None:
        assert list_example_2.mean_len() == 5.0

    # apply
    def test_apply(self, list_example: ListExt) -> None:
        assert list_example.apply(str) == list_example.stringify()

    # unique
    def test_unique(self) -> None:
        assert ListExt([1, 1, 1, 1]).unique() == [1]

    # head
    def test_head(self, list_example: ListExt) -> None:
        assert list_example.head(3) == [3, 8, 5]

    def test_head_2(self, list_example: ListExt) -> None:
        """Max head len"""
        assert list_example.head(100) == list(list_example)

    def test_head_3(self) -> None:
        """Empty list"""
        assert ListExt([]).head(9) == []

    # tail
    def test_tail(self, list_example_2: ListExt) -> None:
        assert list_example_2.tail(2) == ["Tent", "Strong"]

    def test_tail_2(self, list_example_2: ListExt) -> None:
        assert list_example_2.tail(100) == list(list_example_2)

    def test_tail_3(self) -> None:
        """Empty list"""
        assert ListExt([]).tail(9) == []

    # get_random
    def test_get_random(self, list_example_2: ListExt) -> None:
        test = list_example_2.get_random(20)
        assert len(test) == 20

    # flatten
    def test_flatten(self, list_example: ListExt) -> None:
        test = list_example.flatten()
        assert test

    def test_flatten_2(self) -> None:
        test = ListExt([[[[1]]]])
        assert test.flatten(recursive=True) == [1]

    # split chunk
    @pytest.mark.parametrize(
        ["value", "output"],
        [
            (-1, [[1], [1], [1], [1], [1], [1], [1], [1], [1]]),
            (1, [[1], [1], [1], [1], [1], [1], [1], [1], [1]]),
            (2, [[1, 1], [1, 1], [1, 1], [1, 1], [1]]),
            (5, [[1, 1, 1, 1, 1], [1, 1, 1, 1]]),
            (100, [[1, 1, 1, 1, 1, 1, 1, 1, 1]]),
        ],
    )
    def test_split_chunk(self, value: int, output: list[list]) -> None:
        test = ListExt([1, 1, 1, 1, 1, 1, 1, 1, 1])
        assert test.split_chunk(value) == output

    # max item len
    def test_max_item_len(self) -> None:
        ins = ListExt(["test", "longer_test"])
        assert ins.max_item_len() == 11

    def test_max_item_len_2(self) -> None:
        ins = ListExt([["short"], [[["longer_test"]]]])
        assert ins.max_item_len(recursive=True) == 11



"""
Absfuyu: Human
--------------
Human related stuff

Version: 5.17.0
Date updated: 20/11/2025 (dd/mm/yyyy)
"""

# Module level
# ---------------------------------------------------------------------------
__all__ = ["Human", "Person"]


# Library
# ---------------------------------------------------------------------------
from datetime import date, datetime, time, timedelta
from enum import StrEnum
from typing import Self

from absfuyu.core import BaseClass
from absfuyu.dxt import IntExt
from absfuyu.fun import zodiac_sign

# from dateutil.relativedelta import relativedelta


# Class
# ---------------------------------------------------------------------------
class BloodType(StrEnum):
    A_PLUS = "A+"
    A_MINUS = "A-"
    AB_PLUS = "AB+"
    AB_MINUS = "AB-"
    B_PLUS = "B+"
    B_MINUS = "B-"
    O_PLUS = "O+"
    O_MINUS = "O-"
    A = "A"
    AB = "AB"
    B = "B"
    O = "O"
    OTHER = "OTHER"


class Human(BaseClass):
    """
    Basic human data
    """

    def __init__(
        self,
        first_name: str,
        last_name: str | None = None,
        birthday: str | datetime | None = None,
        birth_time: str | None = None,
        gender: bool | None = None,
        height: int | float | None = None,
        weight: int | float | None = None,
        blood_type: BloodType = BloodType.OTHER,
    ) -> None:
        """
        Human instance

        Parameters
        ----------
        first_name : str
            First name

        last_name : str | None, optional
            Last name, by default ``None``

        birthday : str | datetime | None, optional
            Birthday in format: ``yyyy/mm/dd``, by default ``None`` (birthday = today)

        birth_time : str | None, optional
            Birth time in format: ``hh:mm``, by default ``None`` (birthtime = today)

        gender : bool | None, optional
            ``True``: Male; ``False``: Female (biologicaly), by default ``None``

        height : int | float | None, optional
            Height in centimeter (cm), by default ``None``

        weight : int | float | None, optional
            Weight in kilogram (kg), by default ``None``

        blood_type : BloodType, optional
            Blood type, by default ``BloodType.OTHER``
        """

        # Name
        self.first_name = first_name
        self.last_name = last_name
        self.name = (
            f"{self.last_name}, {self.first_name}"
            if self.last_name is not None
            else self.first_name
        )

        # Birthday
        now = datetime.now()
        if birthday is None:
            modified_birthday = now.date()
        elif isinstance(birthday, str):
            for x in ["/", "-"]:
                birthday = birthday.replace(x, "/")
            modified_birthday = datetime.strptime(birthday, "%Y/%m/%d")
        else:
            modified_birthday = birthday

        if birth_time is None:
            modified_birthtime = now.time()
        else:
            birth_time = list(map(int, birth_time.split(":")))  # type: ignore
            modified_birthtime = time(*birth_time)

        self.birthday = date(
            modified_birthday.year, modified_birthday.month, modified_birthday.day
        )
        self.birth_time = modified_birthtime

        self.birth = datetime(
            modified_birthday.year,
            modified_birthday.month,
            modified_birthday.day,
            modified_birthtime.hour,
            modified_birthtime.minute,
        )

        # Others
        self.gender = gender
        self.height = height
        self.weight = weight
        self.blood_type = blood_type

    def __str__(self) -> str:
        class_name = self.__class__.__name__
        return f"{class_name}({str(self.name)})"

    @classmethod
    def JohnDoe(cls) -> Self:
        """
        Dummy Human for test

        Returns
        -------
        Human
            Dummy Human instance
        """
        return cls("John", "Doe", "1980/01/01", "00:00", True, 180, 80, BloodType.O)

    @property
    def is_male(self) -> bool:
        """
        Check if male (biological)

        Returns
        -------
        bool
            - ``True``: Male
            - ``False``: Female
        """
        if self.gender is None:
            raise ValueError("Gender must be defined first")
        return self.gender

    @property
    def age(self) -> float:
        """
        Calculate age based on birthday, precise to birth_time

        Returns
        -------
        float
            Age
        """
        now = datetime.now()
        # age = now - self.birthday
        rdelta = now - self.birth
        # rdelta = relativedelta(now, self.birthday)
        # return round(rdelta.years + rdelta.months / 12, 2)
        return round(rdelta / timedelta(days=365.2425), 2)

    @property
    def is_adult(self) -> bool:
        """
        Check if ``self.age`` >= ``18``

        Returns
        -------
        bool
            If is adult
        """
        return self.age >= 18

    @property
    def bmi(self) -> float:
        r"""
        Body Mass Index (kg/m^2)

        Formula: :math:`\frac{weight (kg)}{height (m)^2}`

        - BMI < 18.5: Skinny
        - 18.5 < BMI < 25: Normal
        - BMI > 30: Obesse

        Returns
        -------
        float
            BMI value
        """
        if self.height is None or self.weight is None:
            raise ValueError("Height and Weight must be defined")

        height_in_meter = self.height / 100
        bmi = self.weight / (height_in_meter**2)
        return round(bmi, 2)

    def update(self, data: dict) -> None:
        """
        Update Human data

        Parameters
        ----------
        data : dict
            Data

        Returns
        -------
        None
        """
        self.__dict__.update(data)


class Person(Human):
    """
    More detailed ``Human`` data
    """

    def __init__(
        self,
        first_name,
        last_name=None,
        birthday=None,
        birth_time=None,
        gender=None,
        height=None,
        weight=None,
        blood_type=BloodType.OTHER,
    ) -> None:
        super().__init__(
            first_name,
            last_name,
            birthday,
            birth_time,
            gender,
            height,
            weight,
            blood_type,
        )
        self.address: str = None  # type: ignore
        self.hometown: str = None  # type: ignore
        self.email: str = None  # type: ignore
        self.phone_number: str = None  # type: ignore
        self.nationality: str = None  # type: ignore
        self.likes: list = None  # type: ignore
        self.hates: list = None  # type: ignore
        self.education: str = None  # type: ignore
        self.occupation: str = None  # type: ignore
        self.personality: str = None  # type: ignore
        self.note: str = None  # type: ignore

    @property
    def zodiac_sign(self) -> str:
        """
        Zodiac sign of ``Person``

        Returns
        -------
        str
            Zodiac sign
        """
        return zodiac_sign(self.birthday.day, self.birthday.month)

    @property
    def zodiac_sign_13(self) -> str:
        """
        Zodiac sign of ``Person`` (13 zodiac signs version)

        Returns
        -------
        str
            Zodiac sign
        """
        return zodiac_sign(self.birthday.day, self.birthday.month, zodiac13=True)

    @property
    def numerology(self) -> int:
        """
        Numerology number of ``Person``

        Returns
        -------
        int
            Numerology number
        """
        temp = f"{self.birthday.year}{self.birthday.month}{self.birthday.day}"
        return IntExt(temp).add_to_one_digit(master_number=True)



"""
Test: Generator

Version: 5.17.0
Date updated: 20/11/2025 (dd/mm/yyyy)
"""

import random
import re
import string
from math import comb

import pytest

from absfuyu.tools.generator import Charset, Generator


class TestGenerator:
    """absfuyu.general.generator.Generator"""

    # Charset
    @pytest.mark.parametrize(
        ["value", "output"],
        [
            (Charset.PRODUCT_KEY, "BCDFGHJKMNPQRTVWXY2346789"),
            (Charset.DEFAULT, string.ascii_letters + string.digits),
        ],
    )
    def test_Charset(self, value: Charset, output: str) -> None:
        assert value == output

    # generate_string
    def test_generate_string(self) -> None:
        """Correct len and a list"""
        temp = Generator.generate_string(
            charset=Charset.DEFAULT,
            size=8,
            times=1,
            unique=False,
            string_type_if_1=False,
        )
        assert isinstance(temp, list) and len(temp[0]) == 8

    def test_generate_string_2(self) -> None:
        """Correct len and a str"""
        temp = Generator.generate_string(
            charset=Charset.DEFAULT,
            size=8,
            times=1,
            unique=False,
            string_type_if_1=True,
        )
        assert isinstance(temp, str) and len(temp) == 8

    def test_generate_string_3(self) -> None:
        """Unique generate"""
        temp = Generator.generate_string(
            charset=Charset.DEFAULT,
            size=2,
            times=600,
            unique=True,
        )
        assert len(temp) == len(list(set(temp)))

    # generate_key
    def test_generate_key(self) -> None:
        """Check if key generate correctly"""
        key = Generator.generate_key(
            charset=Charset.PRODUCT_KEY, letter_per_block=5, number_of_block=5, sep="-"
        )
        key_pattern = r"[A-Z0-9]{5}-[A-Z0-9]{5}-[A-Z0-9]{5}-[A-Z0-9]{5}-[A-Z0-9]{5}"
        check_result = re.search(key_pattern, key)  # Correct default key pattern
        correct_len = len(key) == 29  # Correct default key length
        assert check_result is not None and correct_len

    # generate_check_digit
    @pytest.mark.parametrize(
        ["value", "output"],
        [
            (95489683516944151927, 0),
            (15392479156575151882, 1),
            (74662116892282572844, 2),
            (91274812716671415644, 3),
            (94984564168167844561, 4),
            (94273419372476632513, 5),
            (78985469454121383396, 6),
            (34458526632449856638, 7),
            (95486688921998713381, 8),
            (48981383446354864289, 9),
            ("95489683516944151927", 0),
            ([1, 5, 3, 9, 2, 4, 7, 9, 1, 5, 6, 5, 7, 5, 1, 5, 1, 8, 8, 2], 1),
        ],
    )
    def test_generate_check_digit(self, value: int, output: int) -> None:
        assert Generator.generate_check_digit(value) == output

    # combinations range
    def test_combinations_range(self) -> None:
        min_len = random.randrange(1, 3 + 1)
        max_len = random.randrange(3, 5 + 1)
        random_list = [
            Generator.generate_string(times=1, string_type_if_1=True)
            for _ in range(max_len)
        ]
        # print(min_len, max_len, random_list, comb(max_len, min_len))
        test = Generator.combinations_range(
            random_list, min_len=min_len, max_len=max_len
        )
        len_of_generated_combinations = len(test)
        calculated_total_combinations = sum(
            [comb(max_len, i) for i in range(min_len, max_len + 1)]
        )
        assert len_of_generated_combinations == calculated_total_combinations



"""
Absfuyu: CLI
------------
Custom Argument Parser

Version: 5.17.0
Date updated: 20/11/2025 (dd/mm/yyyy)
"""

# Module level
# ---------------------------------------------------------------------------
__all__ = ["FuyuArgumentParser"]


# Library
# ---------------------------------------------------------------------------
import logging
import sys
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser
from typing import Any, Callable


# Class
# ---------------------------------------------------------------------------
class FuyuArgumentParser(ArgumentParser):
    """
    Usage:
    ------
    >>> import sys
    >>> args = FuyuArgumentParser()
    >>> args.start()
    """

    def __init__(
        self,
        prog: str | None = None,
        description: str | None = None,
        epilog: str | None = None,
        version: str = "0.0.1",
        formatter_class: Any = ArgumentDefaultsHelpFormatter,
        prefix_chars: str = "-",
        argument_default: Any = None,
        conflict_handler: str = "error",
        add_help: bool = True,
        allow_abbrev: bool = True,
        exit_on_error: bool = True,
    ) -> None:
        # Desc
        if description is None:
            description = f"Absfuyu CLI {version}"

        # Default
        self._default_args = ["--help"]

        # Super
        super().__init__(
            prog=prog,
            description=description,
            epilog=epilog,
            formatter_class=formatter_class,
            prefix_chars=prefix_chars,
            argument_default=argument_default,
            conflict_handler=conflict_handler,
            add_help=add_help,
            allow_abbrev=allow_abbrev,
            exit_on_error=exit_on_error,
        )

        # Add version
        self.add_argument("-v", "--version", action="version", version=f"%(prog)s {version}")

        # Add log level
        _ll_val = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
        self.add_argument(
            "-ll",
            "--log-level",
            metavar="LOG_LEVEL",
            dest="log_level",
            choices=_ll_val,
            default="ERROR",
            help=f"Log level: {', '.join(_ll_val)}",
        )

    def _fuyu_set_log_level(self, log_level: str = "ERROR") -> None:
        log_levels = {
            "DEBUG": logging.DEBUG,
            "INFO": logging.INFO,
            "WARNING": logging.WARNING,
            "ERROR": logging.ERROR,
            "CRITICAL": logging.CRITICAL,
        }
        logging.basicConfig(
            level=log_levels[log_level],
            format="[%(asctime)s] [%(module)s] [%(name)s] [%(funcName)s] [%(levelname)-s] %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
        )

    def start(self, default: Callable[[], Any] | None = None):
        """
        Quick start for argument parser

        Parameters
        ----------
        default : Callable[[], Any] | None, optional
            | Default callable to run, by default ``None``
            | Shows help on default behavior (default=None)

        Returns
        -------
        Namespace
            Namespace
        """
        if default is None:
            args = self.parse_args(args=None if sys.argv[1:] else self._default_args)
        else:
            args = self.parse_args(args=None if sys.argv[1:] else default())
        self._fuyu_set_log_level(getattr(args, "log_level", "ERROR"))
        return args



"""
Absfuyu: Json Method
--------------------
``.json`` file handling

Version: 5.17.0
Date updated: 20/11/2025 (dd/mm/yyyy)
"""

# Module level
# ---------------------------------------------------------------------------
__all__ = ["JsonFile"]


# Library
# ---------------------------------------------------------------------------
import json
from pathlib import Path

from absfuyu.core import BaseClass


# Class
# ---------------------------------------------------------------------------
class JsonFile(BaseClass):
    """
    ``.json`` file handling

    Parameters
    ----------
    json_file_location : str | Path
        .json file location

    encoding : str | None, optional
        Data encoding, by default ``"utf-8"``

    indent : int | str | None, optional
        Indentation when export to json file, by default ``4``

    sort_keys : bool, optional
        Sort the keys before export to json file, by default ``True``
    """

    def __init__(
        self,
        json_file_location: str | Path,
        *,
        encoding: str | None = "utf-8",
        indent: int | str | None = 4,
        sort_keys: bool = True,
    ) -> None:
        """
        ``.json`` file handling

        Parameters
        ----------
        json_file_location : str | Path
            .json file location

        encoding : str | None, optional
            Data encoding, by default ``"utf-8"``

        indent : int | str | None, optional
            Indentation when export to json file, by default ``4``

        sort_keys : bool, optional
            Sort the keys before export to json file, by default ``True``
        """

        self.json_file_location = Path(json_file_location)
        self.encoding = encoding
        self.indent = indent
        self.sort_keys = sort_keys
        self.data: dict = {}

    def __str__(self) -> str:
        return f"{self.__class__.__name__}({self.json_file_location.name})"

    def load_json(self) -> dict:
        """
        Load ``.json`` file

        Returns
        -------
        dict
            ``.json`` data
        """
        with open(self.json_file_location, "r", encoding=self.encoding) as file:
            self.data = json.load(file)
        return self.data

    def save_json(self) -> None:
        """Save ``.json`` file"""
        json_data = json.dumps(self.data, indent=self.indent, sort_keys=self.sort_keys)
        with open(self.json_file_location, "w", encoding=self.encoding) as file:
            file.writelines(json_data)

    def update_data(self, data: dict) -> None:
        """
        Update ``.json`` data without save

        Parameters
        ----------
        data : dict
            ``.json`` data
        """
        self.data = data



"""
ABSFUYU CLI
-----------
Config

Version: 5.17.0
Date updated: 20/11/2025 (dd/mm/yyyy)
"""

# Module Package
# ---------------------------------------------------------------------------
__all__ = ["config_group"]


# Library
# ---------------------------------------------------------------------------
import click

from absfuyu.cli.color import COLOR
from absfuyu.config import ABSFUYU_CONFIG


# CLI
# ---------------------------------------------------------------------------
@click.command()
@click.option(
    "--setting",
    "-s",
    type=click.Choice(["luckgod", "install-extra"]),
    help="Toggle on/off selected setting",
)
def toggle(setting: str) -> None:
    """Toggle on/off setting"""

    # Dictionary
    trans = {
        "luckgod": "luckgod-mode",
        "install-extra": "auto-install-extra",
    }  # trans[setting]

    if setting is None:
        click.echo(f"{COLOR['red']}Invalid setting")  # type: ignore
    else:
        ABSFUYU_CONFIG.toggle_setting(trans[setting])
        out = ABSFUYU_CONFIG._get_setting(trans[setting])
        click.echo(f"{COLOR['red']}{out}")


@click.command()
def reset():
    """Reset config to default value"""
    ABSFUYU_CONFIG.reset_config()
    click.echo(f"{COLOR['green']}All settings have been reseted")


@click.group(name="config")
def config_group() -> None:
    """absfuyu configuration"""
    pass


config_group.add_command(toggle)
config_group.add_command(reset)



"""
Absfuyu: Package data
---------------------
Deprecated (but might have some use)

Version: 5.17.0
Date updated: 20/11/2025 (dd/mm/yyyy)
"""

# Library
# ---------------------------------------------------------------------------
import zlib
from ast import literal_eval
from importlib.resources import files, read_binary
from pathlib import Path

from absfuyu.core import BaseClass
from absfuyu.logger import logger

# External Data
# ---------------------------------------------------------------------------
# These are sources of data
_EXTERNAL_DATA = {
    "chemistry.json": "https://raw.githubusercontent.com/Bowserinator/Periodic-Table-JSON/master/PeriodicTableJSON.json",
    "tarot.json": "https://raw.githubusercontent.com/dariusk/corpora/master/data/divination/tarot_interpretations.json",
    "word_list.json": "https://raw.githubusercontent.com/dwyl/english-words/master/words_dictionary.json",
    # "countries.json": "https://raw.githubusercontent.com/dr5hn/countries-states-cities-database/refs/heads/master/json/countries.json",
}


# Deprecated
# ---------------------------------------------------------------------------
DATA_PATH = files("absfuyu.pkg_data")


class _DatMaker(BaseClass):
    """
    Package data ``.dat`` maker/loader

    *Deprecated since version 5.0.0*
    """

    def __init__(self, data_name: str) -> None:
        self.name = data_name

    def _make_dat(self, data: str, name: str | Path):
        """
        :param data: string data
        :param name: name and location of the data
        """
        compressed_data = zlib.compress(str(data).encode(), zlib.Z_BEST_COMPRESSION)
        with open(name, "wb") as file:
            file.write(compressed_data)

    def load_dat_data(self, evaluate: bool = False):
        """
        Load ``.dat`` data from package resource

        :param evaluate: use ``ast.literal_eval()`` to evaluate string data
        :type evaluate: bool
        :returns: Loaded data
        :rtype: Any
        """
        compressed_data = read_binary("absfuyu.pkg_data", self.name)
        data = zlib.decompress(compressed_data).decode()
        # return data
        return literal_eval(data) if evaluate else data

    def update_data(self, new_data: str):
        """
        Update existing data

        :param new_data: Data to be updated
        """
        self._make_dat(data=new_data, name=DATA_PATH.joinpath(self.name))  # type:ignore
        logger.debug("Data updated")


class _ManagePkgData(BaseClass):
    """
    Manage this package data

    *Deprecated since version 5.0.0*
    """

    def __init__(self, pkg_data_loc: str | Path) -> None:
        """
        pkg_data_loc: Package data location
        """
        self.data_loc = Path(pkg_data_loc)

    def __str__(self) -> str:
        return f"{self.__class__.__name__}({self.data_loc.name})"

    def get_data_list(self, *, pattern: str = "*") -> list[Path]:
        """Get a list of data available"""
        excludes = [
            x for x in self.data_loc.glob("*.[pP][yY]")
        ]  # exclude python scripts
        return [
            x for x in self.data_loc.glob(pattern) if x not in excludes and x.is_file()
        ]

    @property
    def data_list(self) -> list[str]:
        """List of available data"""
        return [x.name for x in self.get_data_list()]

    def download_all_data(self):
        """
        Download all external data
        """

        logger.debug("Downloading data...")
        try:
            from absfuyu.util.api import APIRequest

            for data_name, data_link in _EXTERNAL_DATA.items():
                logger.debug(f"Downloading {data_name}...")
                data = APIRequest(data_link, encoding="utf-8")
                data.fetch_data(
                    update=True,
                    json_cache=DATA_PATH.joinpath(data_name),  # type: ignore
                )
                logger.debug(f"Downloading {data_name}...DONE")
            logger.debug("Downloading data...DONE")
        except Exception:
            logger.debug("Downloading data...FAILED")

    def clear_data(self) -> None:
        """Clear data in data list"""
        for x in self.get_data_list():
            x.unlink()



"""
Test: Shape

Version: 5.17.0
Date updated: 20/11/2025 (dd/mm/yyyy)
"""

import math

import pytest

from absfuyu.general.shape import (
    Circle,
    Cube,
    Cuboid,
    Cylinder,
    HemiSphere,
    Hexagon,
    Parallelogram,
    Pentagon,
    Rectangle,
    Rhombus,
    Sphere,
    Square,
    Trapezoid,
    Triangle,
)


@pytest.mark.abs_shape
class TestTriangle:
    def test_initialization_valid(self) -> None:
        """Test valid triangle initialization."""
        triangle = Triangle(3, 4, 5)
        assert triangle.a == 3
        assert triangle.b == 4
        assert triangle.c == 5

    def test_initialization_invalid_negative(self) -> None:
        """Test initialization with negative side lengths."""
        with pytest.raises(ValueError, match="Side lengths must be positive."):
            Triangle(-1, 2, 3)

    def test_initialization_invalid_zero(self) -> None:
        """Test initialization with zero side lengths."""
        with pytest.raises(ValueError, match="Side lengths must be positive."):
            Triangle(0, 1, 2)

    def test_initialization_invalid_triangle_inequality(self) -> None:
        """Test initialization that violates the triangle inequality theorem."""
        with pytest.raises(
            ValueError, match="The provided lengths do not form a valid triangle."
        ):
            Triangle(1, 2, 3)  # 1 + 2 is not greater than 3

    def test_perimeter(self) -> None:
        """Test perimeter calculation."""
        triangle = Triangle(3, 4, 5)
        assert triangle.perimeter() == 12

    def test_area(self) -> None:
        """Test area calculation using Heron's formula."""
        triangle = Triangle(3, 4, 5)
        assert math.isclose(triangle.area(), 6.0)

    def test_is_right_angled(self) -> None:
        """Test right-angled triangle detection."""
        right_triangle = Triangle(3, 4, 5)
        assert right_triangle.is_right_angled() is True

        not_right_triangle = Triangle(2, 2, 3)
        assert not_right_triangle.is_right_angled() is False

    def test_is_equilateral(self) -> None:
        """Test equilateral triangle detection."""
        equilateral_triangle = Triangle(5, 5, 5)
        assert equilateral_triangle.is_equilateral() is True

        not_equilateral_triangle = Triangle(5, 5, 4)
        assert not_equilateral_triangle.is_equilateral() is False

    def test_is_isosceles(self) -> None:
        """Test isosceles triangle detection."""
        isosceles_triangle = Triangle(5, 5, 3)
        assert isosceles_triangle.is_isosceles() is True

        not_isosceles_triangle = Triangle(3, 4, 5)
        assert not_isosceles_triangle.is_isosceles() is False

    @pytest.mark.parametrize(
        ["a", "b", "c", "result"],
        [
            (5, 5, 5, "equilateral"),
            (5, 5, 3, "isosceles"),
            (1, 1, math.sqrt(2), "right-angled isosceles"),
            (3, 4, 5, "right-angled"),
            (3, 4, 6, "scalene"),
        ],
    )
    def test_triangle_type(
        self, a: int | float, b: int | float, c: int | float, result: str
    ) -> None:
        """Test the type of the triangle."""
        triangle = Triangle(a, b, c)
        assert triangle.triangle_type() == result

    def test_is_acute(self) -> None:
        """Test acute triangle detection."""
        acute_triangle = Triangle(4, 5, 6)  # Angles <90 degrees
        assert acute_triangle.is_acute() is True

        obtuse_triangle = Triangle(3, 4, 6)  # One angle >90 degrees
        assert obtuse_triangle.is_acute() is False

    def test_is_obtuse(self) -> None:
        """Test obtuse triangle detection."""
        obtuse_triangle = Triangle(3, 4, 6)  # One angle >90 degrees
        assert obtuse_triangle.is_obtuse() is True

        acute_triangle = Triangle(4, 5, 6)  # Angles <90 degrees
        assert acute_triangle.is_obtuse() is False

    def test_get_angles(self) -> None:
        """Test angle calculation."""
        triangle = Triangle(3, 4, 5)

        angles = [round(x, 2) for x in triangle.get_angles()]

        # Check if angles are approximately correct (in degrees)
        expected_angles = [36.87, 53.13, 90.0]

        for angle in angles:
            assert any(
                math.isclose(angle, expected_angle, abs_tol=1e-4)
                for expected_angle in expected_angles
            )

    def test_scale(self) -> None:
        """Test scaling of the triangle."""
        triangle = Triangle(3, 4, 5)

        # Scale by a factor of 2
        triangle.scale(2)

        assert triangle.a == 6
        assert triangle.b == 8
        assert triangle.c == 10

    def test_scale_invalid_factor(self) -> None:
        """Test scaling with an invalid factor."""
        triangle = Triangle(3, 4, 5)

        with pytest.raises(ValueError) as excinfo:
            triangle.scale(-1)
        assert str(excinfo.value) == "Scaling factor must be positive."

    def test_is_similar(self) -> None:
        """Test similarity of triangles."""
        triangle1 = Triangle(3, 4, 5)
        triangle2 = Triangle(6, 8, 10)  # Similar to triangle1 (scaled by factor of two)

        assert triangle1.is_similar(triangle2) is True

        different_triangle = Triangle(7, 8, 9)
        assert triangle1.is_similar(different_triangle) is False


@pytest.mark.abs_shape
class TestCircle:
    def test_initialization_valid(self) -> None:
        """Test valid circle initialization."""
        circle = Circle(5)
        assert circle.radius == 5

    def test_initialization_invalid(self) -> None:
        """Test initialization with invalid (non-positive) radius."""
        with pytest.raises(ValueError, match="Radius must be a positive number."):
            Circle(-1)
        with pytest.raises(ValueError, match="Radius must be a positive number."):
            Circle(0)

    def test_diameter(self) -> None:
        """Test diameter calculation."""
        circle = Circle(5)
        assert circle.diameter() == 10

    def test_perimeter(self) -> None:
        """Test perimeter calculation (same as circumference)."""
        circle = Circle(5)
        assert math.isclose(circle.perimeter(), 2 * math.pi * 5)

    def test_circumference(self) -> None:
        """Test circumference calculation."""
        circle = Circle(5)
        assert math.isclose(circle.circumference(), 2 * math.pi * 5)

    def test_area(self) -> None:
        """Test area calculation."""
        circle = Circle(5)
        assert math.isclose(circle.area(), math.pi * 5**2)

    def test_scale_valid(self) -> None:
        """Test scaling with a valid (positive) factor."""
        circle = Circle(5)
        circle.scale(2)
        assert circle.radius == 10

    def test_scale_invalid(self) -> None:
        """Test scaling with an invalid (non-positive) factor."""
        circle = Circle(5)
        with pytest.raises(ValueError, match="Scaling factor must be positive."):
            circle.scale(-1)
        with pytest.raises(ValueError, match="Scaling factor must be positive."):
            circle.scale(0)

    def test_equality(self) -> None:
        """Test equality comparison (__eq__)."""
        circle1 = Circle(5)
        circle2 = Circle(5)
        circle3 = Circle(6)
        assert circle1 == circle2
        assert not (circle1 == circle3)

    def test_less_than(self) -> None:
        """Test less than comparison (__lt__)."""
        circle1 = Circle(5)
        circle2 = Circle(6)
        assert circle1 < circle2
        assert not (circle2 < circle1)
        assert not (circle1 < circle1)  # Not less than itself

    def test_less_than_or_equal(self) -> None:
        """Test less than or equal comparison (__le__)."""
        circle1 = Circle(5)
        circle2 = Circle(6)
        circle3 = Circle(5)
        assert circle1 <= circle2
        assert circle1 <= circle3
        assert not (circle2 <= circle1)

    def test_from_area_valid(self) -> None:
        circle = Circle.from_area(math.pi * 5**2)  # Area of circle with radius 5
        assert math.isclose(circle.radius, 5)
        assert Circle(5).radius == Circle.from_area(Circle(5).area()).radius

    def test_from_area_invalid(self) -> None:
        with pytest.raises(ValueError, match="Area must be positive."):
            Circle.from_area(-10)

    def test_from_circumference_valid(self) -> None:
        circle = Circle.from_circumference(
            2 * math.pi * 5
        )  # Circumference of circle with radius 5
        assert math.isclose(circle.radius, 5)
        assert (
            Circle(5).radius
            == Circle.from_circumference(Circle(5).circumference()).radius
        )

    def test_from_circumference_invalid(self) -> None:
        with pytest.raises(ValueError, match="Circumference must be positive."):
            Circle.from_circumference(-10)


@pytest.mark.abs_shape
class TestSquare:
    def test_valid_side(self) -> None:
        square = Square(5)
        assert square.side == 5

    def test_invalid_side(self) -> None:
        with pytest.raises(ValueError, match="Side length must be a positive number."):
            Square(-5)
        with pytest.raises(ValueError, match="Side length must be a positive number."):
            Square(0)

    def test_perimeter(self) -> None:
        square = Square(5)
        assert square.perimeter() == 20

    def test_area(self) -> None:
        square = Square(5)
        assert square.area() == 25

    def test_diagonal(self) -> None:
        square = Square(5)
        assert square.diagonal() == math.sqrt(2) * 5

    def test_from_perimeter_valid(self) -> None:
        square = Square.from_perimeter(20)
        assert square.side == 5
        assert Square(5).side == Square.from_perimeter(Square(5).perimeter()).side

    def test_from_perimeter_invalid(self) -> None:
        with pytest.raises(ValueError, match="Perimeter must be a positive number."):
            Square.from_perimeter(-20)

    def test_from_area_valid(self) -> None:
        square = Square.from_area(25)
        assert square.side == 5
        assert Square(5).side == Square.from_area(Square(5).area()).side

    def test_from_area_invalid(self) -> None:
        with pytest.raises(ValueError, match="Area must be a positive number."):
            Square.from_area(-25)

    def test_scale_valid(self) -> None:
        """Test scaling with a valid (positive) factor."""
        square = Square(5)
        square.scale(2)
        assert square.side == 10

    def test_scale_invalid(self) -> None:
        """Test scaling with an invalid (non-positive) factor."""
        square = Square(5)
        with pytest.raises(ValueError, match="Scaling factor must be positive."):
            square.scale(-1)
        with pytest.raises(ValueError, match="Scaling factor must be positive."):
            square.scale(0)


@pytest.mark.abs_shape
class TestRectangle:
    def test_initialization_valid(self) -> None:
        """Test valid rectangle initialization."""
        rectangle = Rectangle(5, 3)
        assert rectangle.length == 5
        assert rectangle.width == 3

    def test_initialization_invalid_negative(self) -> None:
        """Test initialization with negative length or width."""
        with pytest.raises(
            ValueError, match="Length and width must be positive numbers."
        ):
            Rectangle(-1, 2)
        with pytest.raises(
            ValueError, match="Length and width must be positive numbers."
        ):
            Rectangle(0, 1)

    def test_initialization_invalid_zero(self) -> None:
        """Test initialization with zero length or width."""
        with pytest.raises(
            ValueError, match="Length and width must be positive numbers."
        ):
            Rectangle(0, 0)

    def test_perimeter(self) -> None:
        """Test perimeter calculation."""
        rectangle = Rectangle(5, 3)
        assert rectangle.perimeter() == 16

    def test_area(self) -> None:
        """Test area calculation."""
        rectangle = Rectangle(5, 3)
        assert rectangle.area() == 15

    def test_diagonal(self) -> None:
        """Test diagonal calculation."""
        rectangle = Rectangle(5, 3)
        assert math.isclose(rectangle.diagonal(), math.sqrt(5**2 + 3**2))

    def test_is_square_true(self) -> None:
        """Test is_square method for a square."""
        square = Rectangle(4, 4)
        assert square.is_square() is True

    def test_is_square_false(self) -> None:
        """Test is_square method for a non-square rectangle."""
        rectangle = Rectangle(5, 3)
        assert rectangle.is_square() is False

    def test_scale_valid(self) -> None:
        """Test scaling with a valid (positive) factor."""
        rectangle = Rectangle(5, 3)
        rectangle.scale(2)

        assert rectangle.length == 10
        assert rectangle.width == 6

    def test_scale_invalid(self) -> None:
        """Test scaling with an invalid (non-positive) factor."""
        rectangle = Rectangle(5, 3)

        with pytest.raises(ValueError, match="Scaling factor must be positive."):
            rectangle.scale(-1)

        with pytest.raises(ValueError, match="Scaling factor must be positive."):
            rectangle.scale(0)


@pytest.mark.abs_shape
class TestPentagon:
    def test_initialization_valid(self) -> None:
        """Test valid pentagon initialization."""
        pentagon = Pentagon(5)
        assert pentagon.side == 5

    def test_initialization_invalid_negative(self) -> None:
        """Test initialization with negative side length."""
        with pytest.raises(ValueError, match="Side length must be a positive number."):
            Pentagon(-1)

    def test_initialization_invalid_zero(self) -> None:
        """Test initialization with zero side length."""
        with pytest.raises(ValueError, match="Side length must be a positive number."):
            Pentagon(0)

    def test_perimeter(self) -> None:
        """Test perimeter calculation."""
        pentagon = Pentagon(5)
        assert pentagon.perimeter() == 25

    def test_area(self) -> None:
        """Test area calculation."""
        pentagon = Pentagon(5)
        expected_area = (5 / 4) * (5**2) / math.tan(math.pi / 5)
        assert math.isclose(pentagon.area(), expected_area)

    def test_apothem(self) -> None:
        """Test apothem calculation."""
        pentagon = Pentagon(5)
        expected_apothem = (pentagon.side / 2) / math.tan(math.pi / 5)
        assert math.isclose(pentagon.apothem(), expected_apothem)

    def test_area2(self) -> None:
        """Test area calculation using the apothem."""
        pentagon = Pentagon(5)
        expected_area2 = 0.5 * pentagon.perimeter() * pentagon.apothem()
        assert math.isclose(pentagon.area2(), expected_area2)

    def test_from_area_valid(self) -> None:
        """
        Test creating a Pentagon instance from a valid area.
        """
        area = 43.01193501  # Area of pentagon with side 5
        pentagon = Pentagon.from_area(area)
        assert math.isclose(pentagon.side, 5)
        assert Pentagon(5).side == Pentagon.from_area(Pentagon(5).area()).side

    def test_from_area_invalid(self) -> None:
        """
        Test creating a Pentagon instance from an invalid (negative) area.
        """
        with pytest.raises(ValueError, match="Area must be positive."):
            Pentagon.from_area(-10)

    def test_from_perimeter_valid(self) -> None:
        """
        Test creating a Pentagon instance from a valid perimeter.
        """
        perimeter = 5 * 5  # Perimeter of pentagon with side 5
        pentagon = Pentagon.from_perimeter(perimeter)
        assert math.isclose(pentagon.side, 5)
        assert Pentagon(5).side == Pentagon.from_perimeter(Pentagon(5).perimeter()).side

    def test_from_perimeter_invalid(self) -> None:
        """
        Test creating a Pentagon instance from an invalid (negative) perimeter.
        """
        with pytest.raises(ValueError, match="Perimeter must be positive."):
            Pentagon.from_perimeter(-10)


@pytest.mark.abs_shape
class TestHexagon:
    def test_initialization_valid(self) -> None:
        """Test valid hexagon initialization."""
        hexagon = Hexagon(5)
        assert hexagon.side == 5

    def test_initialization_invalid_negative(self) -> None:
        """Test initialization with negative side length."""
        with pytest.raises(ValueError, match="Side length must be a positive number."):
            Hexagon(-1)

    def test_initialization_invalid_zero(self) -> None:
        """Test initialization with zero side length."""
        with pytest.raises(ValueError, match="Side length must be a positive number."):
            Hexagon(0)

    def test_perimeter(self) -> None:
        """Test perimeter calculation."""
        hexagon = Hexagon(5)
        assert hexagon.perimeter() == 30

    def test_area(self) -> None:
        """Test area calculation."""
        hexagon = Hexagon(5)
        expected_area = (3 * math.sqrt(3) / 2) * (5**2)
        assert math.isclose(hexagon.area(), expected_area)

    def test_apothem(self) -> None:
        """Test apothem calculation."""
        hexagon = Hexagon(5)
        expected_apothem = hexagon.side / (2 * math.tan(math.pi / 6))
        assert math.isclose(hexagon.apothem(), expected_apothem)

    def test_from_area_valid(self) -> None:
        """Test creating a Hexagon instance from a valid area."""
        area = (3 * math.sqrt(3) / 2) * (5**2)  # Area of hexagon with side length 5
        hexagon = Hexagon.from_area(area)
        assert math.isclose(hexagon.side, 5)
        assert Hexagon(5).side == Hexagon.from_area(Hexagon(5).area()).side

    def test_from_area_invalid(self) -> None:
        """Test creating a Hexagon instance from an invalid (negative) area."""
        with pytest.raises(ValueError, match="Area must be positive."):
            Hexagon.from_area(-10)

    def test_from_perimeter_valid(self) -> None:
        """Test creating a Hexagon instance from a valid perimeter."""
        perimeter = 6 * 5  # Perimeter of hexagon with side length 5
        hexagon = Hexagon.from_perimeter(perimeter)
        assert math.isclose(hexagon.side, 5)
        assert Hexagon(5).side == Hexagon.from_perimeter(Hexagon(5).perimeter()).side

    def test_from_perimeter_invalid(self) -> None:
        """Test creating a Hexagon instance from an invalid (negative) perimeter."""
        with pytest.raises(ValueError, match="Perimeter must be positive."):
            Hexagon.from_perimeter(-10)


@pytest.mark.abs_shape
class TestParallelogram:
    def test_initialization_valid(self) -> None:
        """Test valid parallelogram initialization."""
        parallelogram = Parallelogram(base=5, height=3, a=4)
        assert parallelogram.base == 5
        assert parallelogram.height == 3
        assert parallelogram.a == 4
        assert parallelogram.phi is None  # phi wasn't specified

    def test_initialization_invalid_base(self) -> None:
        """Test initialization with invalid base."""
        with pytest.raises(ValueError, match="Base must be a positive number."):
            Parallelogram(base=-1, height=3, a=4)
        with pytest.raises(ValueError, match="Base must be a positive number."):
            Parallelogram(base=0, height=3, a=4)

    def test_initialization_invalid_height(self) -> None:
        """Test initialization with invalid height."""
        with pytest.raises(ValueError, match="Height must be a positive number."):
            Parallelogram(base=5, height=-1, a=4)
        with pytest.raises(ValueError, match="Height must be a positive number."):
            Parallelogram(base=5, height=0, a=4)

    def test_initialization_invalid_a(self) -> None:
        """Test initialization with invalid side a."""
        with pytest.raises(ValueError, match="Side 'a' must be a positive number."):
            Parallelogram(base=5, height=3, a=-1)
        with pytest.raises(ValueError, match="Side 'a' must be a positive number."):
            Parallelogram(base=5, height=3, a=0)

    def test_initialization_invalid_phi(self) -> None:
        """Test initialization with invalid angle phi."""
        with pytest.raises(
            ValueError, match="Angle 'phi' must be between 0 and 180 degrees."
        ):
            Parallelogram(base=5, height=3, phi=-1)
        with pytest.raises(
            ValueError, match="Angle 'phi' must be between 0 and 180 degrees."
        ):
            Parallelogram(base=5, height=3, phi=0)
        with pytest.raises(
            ValueError, match="Angle 'phi' must be between 0 and 180 degrees."
        ):
            Parallelogram(base=5, height=3, phi=180)
        with pytest.raises(
            ValueError, match="Angle 'phi' must be between 0 and 180 degrees."
        ):
            Parallelogram(base=5, height=3, phi=181)

    def test_perimeter_with_a(self) -> None:
        """Test perimeter calculation with side a provided."""
        parallelogram = Parallelogram(base=5, height=3, a=4)
        assert parallelogram.perimeter() == 18

    def test_perimeter_with_phi(self) -> None:
        """Test perimeter calculation with angle phi provided."""
        parallelogram = Parallelogram(base=5, height=3, phi=60)
        expected_perimeter = 2 * (5 + 3 / math.sin(math.radians(60)))
        assert math.isclose(parallelogram.perimeter(), expected_perimeter)

    def test_perimeter_invalid(self) -> None:
        """Test perimeter calculation when neither a nor phi is provided."""
        parallelogram = Parallelogram(base=5, height=3)
        with pytest.raises(ValueError, match="Side a or phi must be provided"):
            parallelogram.perimeter()

    def test_area(self) -> None:
        """Test area calculation."""
        parallelogram = Parallelogram(base=5, height=3)
        assert parallelogram.area() == 15


@pytest.mark.abs_shape
class TestRhombus:
    def test_initialization_valid(self) -> None:
        """Test valid rhombus initialization."""
        rhombus = Rhombus(6, 8)
        assert rhombus.d1 == 6
        assert rhombus.d2 == 8

    def test_initialization_invalid_negative(self) -> None:
        """Test initialization with negative diagonal lengths."""
        with pytest.raises(ValueError, match="Diagonal d1 must be a positive number."):
            Rhombus(-1, 8)

        with pytest.raises(ValueError, match="Diagonal d2 must be a positive number."):
            Rhombus(6, -1)

    def test_initialization_invalid_zero(self) -> None:
        """Test initialization with zero diagonal lengths."""
        with pytest.raises(ValueError, match="Diagonal d1 must be a positive number."):
            Rhombus(0, 8)

        with pytest.raises(ValueError, match="Diagonal d2 must be a positive number."):
            Rhombus(6, 0)

    def test_perimeter(self) -> None:
        """Test perimeter calculation."""
        rhombus = Rhombus(6, 8)
        expected_side = rhombus.side()
        assert rhombus.perimeter() == 4 * expected_side

    def test_area(self) -> None:
        """Test area calculation."""
        rhombus = Rhombus(6, 8)
        expected_area = (rhombus.d1 * rhombus.d2) / 2
        assert rhombus.area() == expected_area

    def test_side_calculation(self) -> None:
        """Test side length calculation."""
        rhombus = Rhombus(6, 8)
        expected_side = math.sqrt((rhombus.d1 / 2) ** 2 + (rhombus.d2 / 2) ** 2)
        assert math.isclose(rhombus.side(), expected_side)


@pytest.mark.abs_shape
class TestTrapezoid:
    def test_initialization_valid(self):
        """Test valid trapezoid initialization."""
        trapezoid = Trapezoid(a=5, b=3, c=4, d=6, h=2)
        assert trapezoid.a == 5
        assert trapezoid.b == 3
        assert trapezoid.c == 4
        assert trapezoid.d == 6
        assert trapezoid.h == 2

    def test_initialization_invalid_base_a(self) -> None:
        """Test initialization with invalid base a."""
        with pytest.raises(ValueError, match="Base 'a' must be a positive number."):
            Trapezoid(a=-1, b=3, c=4, d=6, h=2)

        with pytest.raises(ValueError, match="Base 'a' must be a positive number."):
            Trapezoid(a=0, b=3, c=4, d=6, h=2)

    def test_initialization_invalid_base_b(self) -> None:
        """Test initialization with invalid base b."""
        with pytest.raises(ValueError, match="Base 'b' must be a positive number."):
            Trapezoid(a=5, b=-1, c=4, d=6, h=2)

        with pytest.raises(ValueError, match="Base 'b' must be a positive number."):
            Trapezoid(a=5, b=0, c=4, d=6, h=2)

    def test_initialization_invalid_side_c(self) -> None:
        """Test initialization with invalid side c."""
        with pytest.raises(ValueError, match="Side 'c' must be a positive number."):
            Trapezoid(a=5, b=3, c=-1, d=6, h=2)

    def test_initialization_invalid_side_d(self) -> None:
        """Test initialization with invalid side d."""
        with pytest.raises(ValueError, match="Side 'd' must be a positive number."):
            Trapezoid(a=5, b=3, c=4, d=-1, h=2)

    def test_initialization_invalid_height(self) -> None:
        """Test initialization with invalid height."""
        with pytest.raises(ValueError, match="Height 'h' must be a positive number."):
            Trapezoid(a=5, b=3, c=4, d=6, h=-1)

        with pytest.raises(ValueError, match="Height 'h' must be a positive number."):
            Trapezoid(a=5, b=3, c=4, d=6, h=0)

    def test_perimeter_with_sides(self) -> None:
        """Test perimeter calculation when sides c and d are provided."""
        trapezoid = Trapezoid(a=5, b=3, c=4, d=6)
        assert trapezoid.perimeter() == 18  # 5 + 3 + 4 + 6

    def test_perimeter_invalid(self) -> None:
        """Test perimeter calculation when sides c and d are not provided."""
        trapezoid = Trapezoid(a=5, b=3)
        with pytest.raises(
            ValueError, match="'c' and 'd' must be provided to calculate perimeter."
        ):
            trapezoid.perimeter()

    def test_area(self) -> None:
        """Test area calculation."""
        trapezoid = Trapezoid(a=5, b=3, h=2)
        expected_area = (trapezoid.a + trapezoid.b) * trapezoid.h / 2
        assert trapezoid.area() == expected_area

    def test_area_invalid(self) -> None:
        """Test area calculation when height is not provided."""
        trapezoid = Trapezoid(a=5, b=3)
        with pytest.raises(ValueError, match="'h' must be provided to calculate area."):
            trapezoid.area()


@pytest.mark.abs_shape
class TestCube:
    def test_initialization_valid(self) -> None:
        """Test valid cube initialization."""
        cube = Cube(3)
        assert cube.side == 3

    def test_initialization_invalid_negative(self) -> None:
        """Test initialization with negative side length."""
        with pytest.raises(ValueError, match="Side length must be a positive number."):
            Cube(-1)

    def test_initialization_invalid_zero(self) -> None:
        """Test initialization with zero side length."""
        with pytest.raises(ValueError, match="Side length must be a positive number."):
            Cube(0)

    def test_surface_area(self) -> None:
        """Test surface area calculation."""
        cube = Cube(3)
        expected_surface_area = 6 * (3**2)
        assert cube.surface_area() == expected_surface_area

    def test_volume(self) -> None:
        """Test volume calculation."""
        cube = Cube(3)
        expected_volume = 3**3
        assert cube.volume() == expected_volume

    def test_surface_area_side(self) -> None:
        """Test surface area of one side calculation."""
        cube = Cube(3)
        expected_surface_area_side = 4 * (3**2)
        assert cube.surface_area_side() == expected_surface_area_side

    def test_face_area(self) -> None:
        """Test face area calculation."""
        cube = Cube(3)
        expected_face_area = 3**2
        assert cube.face_area() == expected_face_area

    def test_diagonal(self) -> None:
        """Test diagonal calculation."""
        cube = Cube(3)
        expected_diagonal = 3 * math.sqrt(3)
        assert math.isclose(cube.diagonal(), expected_diagonal)

    def test_scale_valid(self) -> None:
        """Test scaling with a valid (positive) factor."""
        cube = Cube(5)
        cube.scale(2)
        assert cube.side == 10

    def test_scale_invalid(self) -> None:
        """Test scaling with an invalid (non-positive) factor."""
        cube = Cube(5)
        with pytest.raises(ValueError, match="Scaling factor must be positive."):
            cube.scale(-1)
        with pytest.raises(ValueError, match="Scaling factor must be positive."):
            cube.scale(0)

    def test_from_surface_area_valid(self) -> None:
        """Test creating a Cube instance from a valid surface area."""
        surface_area = 6 * (5**2)  # Surface area of a cube with side length 5
        cube = Cube.from_surface_area(surface_area)
        assert math.isclose(cube.side, 5)
        assert Cube(5).side == Cube.from_surface_area(Cube(5).surface_area()).side

    def test_from_surface_area_invalid(self) -> None:
        """Test creating a Cube instance from an invalid (negative) surface area."""
        with pytest.raises(ValueError, match="Surface area must be positive."):
            Cube.from_surface_area(-10)

    def test_from_volume_valid(self) -> None:
        """Test creating a Cube instance from a valid volume."""
        volume = 5**3  # Volume of a cube with side length 5
        cube = Cube.from_volume(volume)
        assert math.isclose(cube.side, 5)
        assert Cube(5).side == Cube.from_volume(Cube(5).volume()).side

    def test_from_volume_invalid(self) -> None:
        """Test creating a Cube instance from an invalid (negative) volume."""
        with pytest.raises(ValueError, match="Volume must be positive."):
            Cube.from_volume(-10)


@pytest.mark.abs_shape
class TestCuboid:
    def test_valid_dimensions(self) -> None:
        cuboid = Cuboid(length=5, width=4, height=3)
        assert cuboid.length == 5
        assert cuboid.width == 4
        assert cuboid.height == 3

    def test_invalid_dimensions(self) -> None:
        with pytest.raises(ValueError, match="Length must be a positive number."):
            Cuboid(length=-5, width=4, height=3)

        with pytest.raises(ValueError, match="Width must be a positive number."):
            Cuboid(length=5, width=-4, height=3)

        with pytest.raises(ValueError, match="Height must be a positive number."):
            Cuboid(length=5, width=4, height=-3)

    def test_zero_dimensions(self) -> None:
        with pytest.raises(ValueError, match="Length must be a positive number."):
            Cuboid(length=0, width=4, height=3)
        with pytest.raises(ValueError, match="Width must be a positive number."):
            Cuboid(length=5, width=0, height=3)
        with pytest.raises(ValueError, match="Height must be a positive number."):
            Cuboid(length=5, width=4, height=0)

    def test_surface_area(self) -> None:
        cuboid = Cuboid(length=5, width=4, height=3)
        assert cuboid.surface_area() == 94

    def test_surface_area_side(self) -> None:
        cuboid = Cuboid(length=5, width=4, height=3)
        assert cuboid.surface_area_side() == 54

    def test_volume(self) -> None:
        cuboid = Cuboid(length=5, width=4, height=3)
        assert cuboid.volume() == 60

    def test_diagonal(self) -> None:
        cuboid = Cuboid(length=5, width=4, height=3)
        assert cuboid.diagonal() == math.sqrt(5**2 + 4**2 + 3**2)

    def test_scale_valid(self) -> None:
        cuboid = Cuboid(length=5, width=4, height=3)
        cuboid.scale(2)
        assert cuboid.length == 10
        assert cuboid.width == 8
        assert cuboid.height == 6

    def test_scale_invalid(self) -> None:
        cuboid = Cuboid(length=5, width=4, height=3)
        with pytest.raises(ValueError, match="Scaling factor must be positive."):
            cuboid.scale(-2)


@pytest.mark.abs_shape
class TestSphere:
    """Test suite for the Sphere class."""

    def test_initialization_valid(self) -> None:
        """Test initializing Sphere with a valid radius."""
        sphere = Sphere(5)
        assert sphere.radius == 5

    def test_initialization_invalid(self) -> None:
        """Test initializing Sphere with an invalid (negative) radius."""
        with pytest.raises(ValueError, match="Radius must be a positive number."):
            Sphere(-5)

    def test_surface_area(self) -> None:
        """Test calculating the surface area of the sphere."""
        sphere = Sphere(5)
        expected_surface_area = 4 * math.pi * (5**2)
        assert math.isclose(sphere.surface_area(), expected_surface_area)

    def test_volume(self) -> None:
        """Test calculating the volume of the sphere."""
        sphere = Sphere(5)
        expected_volume = (4 / 3) * math.pi * (5**3)
        assert math.isclose(sphere.volume(), expected_volume)

    def test_from_volume_valid(self) -> None:
        """Test creating a Sphere instance from a valid volume."""
        volume = (4 / 3) * math.pi * (5**3)  # Volume of sphere with radius 5
        sphere = Sphere.from_volume(volume)
        assert math.isclose(sphere.radius, 5)
        assert Sphere(5).radius == Sphere.from_volume(Sphere(5).volume()).radius

    def test_from_volume_invalid(self) -> None:
        """Test creating a Sphere instance from an invalid (negative) volume."""
        with pytest.raises(ValueError, match="Volume must be a positive number."):
            Sphere.from_volume(-10)

    def test_from_surface_area_valid(self) -> None:
        """Test creating a Sphere instance from valid surface area."""
        surface_area = 4 * math.pi * (5**2)  # Surface area of sphere with radius 5
        sphere = Sphere.from_surface_area(surface_area)
        assert math.isclose(sphere.radius, 5)
        assert (
            Sphere(5).radius
            == Sphere.from_surface_area(Sphere(5).surface_area()).radius
        )

    def test_from_surface_area_invalid(self) -> None:
        """Test creating a Sphere instance from an invalid (negative) surface area."""
        with pytest.raises(ValueError, match="Surface area must be a positive number."):
            Sphere.from_surface_area(-10)


@pytest.mark.abs_shape
class TestHemiSphere:
    """
    Test suite for the HemiSphere class.
    """

    def test_initialization_valid(self) -> None:
        """
        Test initializing HemiSphere with a valid radius.
        """
        hemisphere = HemiSphere(5)
        assert hemisphere.radius == 5

    def test_initialization_invalid(self) -> None:
        """
        Test initializing HemiSphere with an invalid (negative) radius.
        """
        with pytest.raises(ValueError, match="Radius must be a positive number."):
            HemiSphere(-5)

    def test_surface_area(self) -> None:
        """
        Test calculating the surface area of the hemisphere.
        """
        hemisphere = HemiSphere(5)
        expected_surface_area = 3 * math.pi * 5**2
        assert math.isclose(hemisphere.surface_area(), expected_surface_area)

    def test_surface_area_curved(self) -> None:
        """
        Test calculating the curved surface area of the hemisphere.
        """
        hemisphere = HemiSphere(5)
        expected_surface_area_curved = 2 * math.pi * 5**2
        assert math.isclose(
            hemisphere.surface_area_curved(), expected_surface_area_curved
        )

    def test_surface_area_base(self) -> None:
        """
        Test calculating the base surface area of the hemisphere
        """
        hemisphere = HemiSphere(5)
        expected_surface_area_base = math.pi * 5**2
        assert math.isclose(hemisphere.surface_area_base(), expected_surface_area_base)

    def test_volume(self) -> None:
        """
        Test calculating the volume of the hemisphere.
        """
        hemisphere = HemiSphere(5)
        expected_volume = (2 / 3) * math.pi * 5**3
        assert math.isclose(hemisphere.volume(), expected_volume)

    def test_from_volume_valid(self) -> None:
        """
        Test creating a HemiSphere instance from a valid volume.
        """
        volume = (2 / 3) * math.pi * 5**3
        hemisphere = HemiSphere.from_volume(volume)
        assert math.isclose(hemisphere.radius, 5)
        assert (
            HemiSphere(5).radius
            == HemiSphere.from_volume(HemiSphere(5).volume()).radius
        )

    def test_from_volume_invalid(self) -> None:
        """
        Test creating a HemiSphere instance from an invalid (negative) volume.
        """
        with pytest.raises(ValueError, match="Volume must be positive."):
            HemiSphere.from_volume(-10)

    def test_from_surface_area_valid(self) -> None:
        """
        Test creating a HemiSphere instance from a valid surface area.
        """
        surface_area = 3 * math.pi * 5**2
        hemisphere = HemiSphere.from_surface_area(surface_area)
        assert math.isclose(hemisphere.radius, 5)
        assert (
            HemiSphere(5).radius
            == HemiSphere.from_surface_area(HemiSphere(5).surface_area()).radius
        )

    def test_from_surface_area_invalid(self) -> None:
        """
        Test creating a HemiSphere instance from an invalid (negative) surface area.
        """
        with pytest.raises(ValueError, match="Surface area must be positive."):
            HemiSphere.from_surface_area(-10)


@pytest.mark.abs_shape
class TestCylinder:
    """Test suite for the Cylinder class."""

    def test_initialization_valid(self) -> None:
        """Test initializing Cylinder with valid radius and height."""
        cylinder = Cylinder(5, 10)
        assert cylinder.radius == 5
        assert cylinder.height == 10

    def test_initialization_invalid_radius(self) -> None:
        """Test initializing Cylinder with an invalid (negative) radius."""
        with pytest.raises(ValueError, match="Radius must be a positive number."):
            Cylinder(-5, 10)

    def test_initialization_invalid_height(self) -> None:
        """Test initializing Cylinder with an invalid (negative) height."""
        with pytest.raises(ValueError, match="Height must be a positive number."):
            Cylinder(5, -10)

    def test_surface_area(self) -> None:
        """Test calculating the total surface area of the cylinder."""
        cylinder = Cylinder(5, 10)
        expected_surface_area = 2 * math.pi * 5 * (5 + 10)
        assert math.isclose(cylinder.surface_area(), expected_surface_area)

    def test_surface_area_curved(self) -> None:
        """Test calculating the curved surface area of the cylinder."""
        cylinder = Cylinder(5, 10)
        expected_curved_surface_area = 2 * math.pi * 5 * 10
        assert math.isclose(
            cylinder.surface_area_curved(), expected_curved_surface_area
        )

    def test_volume(self) -> None:
        """Test calculating the volume of the cylinder."""
        cylinder = Cylinder(5, 10)
        expected_volume = math.pi * (5**2) * 10
        assert math.isclose(cylinder.volume(), expected_volume)



"""
Test: Extra

Version: 5.17.0
Date updated: 20/11/2025 (dd/mm/yyyy)
"""

# Library
# ---------------------------------------------------------------------------
import pytest

from absfuyu import extra as ext


def test_ext_load():
    assert ext.is_loaded() is True



"""
Test: Fun

Version: 5.17.0
Date updated: 20/11/2025 (dd/mm/yyyy)
"""

import pytest

from absfuyu.fun import human_year_to_dog_year, zodiac_sign
from absfuyu.fun.tarot import Tarot, TarotCard


@pytest.mark.abs_fun
class TestFunZodiac:
    """absfuyu.fun.zodiac_sign()"""

    def test_zodiac(self) -> None:
        assert zodiac_sign(1, 1) == "Capricorn (E)"

    def test_zodiac_2(self) -> None:
        assert zodiac_sign(1, 1, zodiac13=True) == "Sagittarius"


@pytest.mark.abs_fun
class TestFunHumanYearToDogYear:
    """absfuyu.fun.human_year_to_dog_year()"""

    def test_first_year(self) -> None:
        assert human_year_to_dog_year(1) == 15

    def test_second_year(self) -> None:
        assert human_year_to_dog_year(2) == 24

    def test_small_dog_after_second_year(self) -> None:
        assert human_year_to_dog_year(3, is_small=True) == 28
        assert human_year_to_dog_year(10, is_small=True) == 56

    def test_large_dog_after_second_year(self) -> None:
        assert human_year_to_dog_year(3, is_small=False) == 29
        assert human_year_to_dog_year(10, is_small=False) == 64

    def test_float_input(self) -> None:
        assert human_year_to_dog_year(1.5) == 19.5
        assert human_year_to_dog_year(2.5) == 26.0
        assert human_year_to_dog_year(3.5, is_small=True) == 30
        assert human_year_to_dog_year(3.5, is_small=False) == 31.5

    def test_negative_input(self) -> None:
        with pytest.raises(ValueError) as excinfo:
            human_year_to_dog_year(-1)
        assert "Value must be positive" in str(excinfo.value)

    def test_zero_input(self) -> None:
        with pytest.raises(ValueError) as excinfo:
            human_year_to_dog_year(-0.1)
        assert "Value must be positive" in str(excinfo.value)


@pytest.mark.abs_fun
class TestFunTarot:
    """absfuyu.fun.tarot.Tarot"""

    def test_tarot(self) -> None:
        assert isinstance(Tarot().random_card(), TarotCard)



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import os
import sys

sys.path.insert(0, os.path.abspath(".."))

from absfuyu import __author__, __title__, __version__

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = f"{__title__}"
copyright = f"2022-%Y, {__author__}"
author = __author__
version = __version__
release = version

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

needs_sphinx = "8.1"

extensions = [
    "sphinx.ext.autodoc",  # include documentation
    "sphinx.ext.autosummary",
    "sphinx.ext.coverage",  # docs coverage
    "sphinx.ext.doctest",  # test docs
    "sphinx.ext.duration",
    # "sphinx.ext.extlinks", # short link
    "sphinx.ext.githubpages",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",  # This shows source code
    "sphinx_copybutton",
    "myst_parser",  # Markdown support
]

templates_path = [
    "_templates"
]

exclude_patterns = [
    "_build", "Thumbs.db", ".DS_Store"
]

language = "en"

source_suffix = {
    ".rst": "restructuredtext",
    ".txt": "markdown",
    ".md": "markdown",
}


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output


html_title = f"{__title__.upper()}'S DOCUMENTATION v{__version__}"
html_short_title = f"{__title__} v{__version__}"
# html_logo = "https://github.com/AbsoluteWinter/AbsoluteWinter.github.io/blob/75568549d30f447a66baac8ecc3b5e6b49a67d18/absfuyu/images/LOGO.png?raw=true"

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = False

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
html_show_sphinx = False

html_sidebars = {
    "**": [
        "versioning.html",
    ],
}

# -- Extension configuration -------------------------------------------------

autodoc_member_order = "bysource"


#  -- sphinx-copybutton configuration ----------------------------------------
copybutton_prompt_text = (
    r">>> |\.\.\. |> |\$ |\# | In \[\d*\]: | {2,5}\.\.\.: | {5,8}: "
)
copybutton_prompt_is_regexp = True
copybutton_remove_prompts = True



"""
Absfuyu: Core
-------------
Decorator

Version: 5.17.0
Date updated: 20/11/2025 (dd/mm/yyyy)
"""

# Module Package
# ---------------------------------------------------------------------------
__all__ = [
    "dummy_decorator",
    "dummy_decorator_with_args",
    "add_subclass_methods_decorator",
]


# Library
# ---------------------------------------------------------------------------
from collections.abc import Callable
from functools import wraps
from typing import ParamSpec, TypeVar, cast, overload

# Type
# ---------------------------------------------------------------------------
P = ParamSpec("P")  # Parameter type
R = TypeVar("R")  # Return type - Can be anything
T = TypeVar("T", bound=type)  # Type type - Can be any subtype of `type`


# Decorator
# ---------------------------------------------------------------------------
@overload
def dummy_decorator(obj: T) -> T: ...
@overload
def dummy_decorator(obj: Callable[P, R]) -> Callable[P, R]: ...
def dummy_decorator(obj: Callable[P, R] | T) -> Callable[P, R] | T:
    """
    This is a decorator that does nothing. Normally used as a placeholder
    """
    if isinstance(obj, type):
        return obj

    @wraps(obj)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
        return obj(*args, **kwargs)

    return wrapper


def dummy_decorator_with_args(*args, **kwargs):
    """
    This is a decorator with args and kwargs that does nothing. Normally used as a placeholder
    """

    @overload
    def decorator(obj: T) -> T: ...

    @overload
    def decorator(obj: Callable[P, R]) -> Callable[P, R]: ...

    def decorator(obj: Callable[P, R] | T) -> Callable[P, R] | T:
        if isinstance(obj, type):
            return obj

        @wraps(obj)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
            return obj(*args, **kwargs)

        return wrapper

    return decorator


def add_subclass_methods_decorator(cls: T) -> T:
    """
    Class decorator replace the ``__init_subclass__`` method.

    This method populates a dictionary with subclass names as keys
    and their available methods as values.

    - Create 2 class attributes: ``_METHOD_INCLUDE`` (bool) and ``SUBCLASS_METHODS`` (dict[str, list[str]])
    - Automatically add subclass methods to class variable: ``SUBCLASS_METHODS``
    - Set class attribute ``_METHOD_INCLUDE`` to ``False`` to exclude from ``SUBCLASS_METHODS``

    Example:
    --------
    >>> # Normal behavior
    >>> @add_subclass_methods_decorator
    >>> class TestParent: ...
    >>> class TestChild(TestParent):
    ...     def method1(self): ...
    >>> TestChild.SUBCLASS_METHODS
    {'__main__.TestChild': ['method1']}

    >>> # Hidden from ``SUBCLASS_METHODS``
    >>> @add_subclass_methods_decorator
    >>> class TestParent: ...
    >>> class TestChildHidden(TestParent):
    ...     _METHOD_INCLUDE = False
    ...     def method1(self): ...
    >>> TestChildHidden.SUBCLASS_METHODS
    {}
    """

    # Check for class
    if not isinstance(cls, type):
        raise ValueError("Object is not a class")

    class AutoSubclassMixin:
        _METHOD_INCLUDE: bool = True  # Include in SUBCLASS_METHODS
        SUBCLASS_METHODS: dict[str, list[str]] = {}

        def __init_subclass__(cls, *args, **kwargs) -> None:
            """
            This create a dictionary with:
            - key   (str)      : Subclass
            - value (list[str]): List of available methods
            """
            super().__init_subclass__(*args, **kwargs)

            if cls._METHOD_INCLUDE and not any(
                [x.endswith(cls.__name__) for x in cls.SUBCLASS_METHODS.keys()]
            ):
                methods_list: list[str] = [
                    k for k, v in cls.__dict__.items() if callable(v)
                ]
                if len(methods_list) > 0:
                    name = f"{cls.__module__}.{cls.__name__}"
                    cls.SUBCLASS_METHODS.update({name: sorted(methods_list)})

    return cast(T, type("AutoSubclass", (AutoSubclassMixin, cls), {}))



"""
Absfuyu: Audio
--------------
Audio convert, lossless checker

Version: 5.17.0
Date updated: 20/11/2025 (dd/mm/yyyy)
"""

# Module level
# ---------------------------------------------------------------------------
__all__ = ["StatusCode", "ResultStatus"]

# Library
# ---------------------------------------------------------------------------
from enum import StrEnum
from pathlib import Path

# Class Enum, Result
# ---------------------------------------------------------------------------
try:
    from rich import print

    class StatusCode(StrEnum):
        OK = "[bold green]OK[/]"
        SKIP = "[bold yellow]SKIP[/]"
        ERROR = "[bold red]ERROR[/]"
        LOSSLESS = "[bold green]LOSSLESS[/]"
        NOT_LOSSLESS = "[bold red]NOT LOSSLESS[/]"
        HIRES = "[bold blue]HIRES[/]"

except ImportError:

    class StatusCode(StrEnum):
        OK = "OK"
        SKIP = "SKIP"
        ERROR = "ERROR"
        LOSSLESS = "LOSSLESS"
        NOT_LOSSLESS = "NOT LOSSLESS"
        HIRES = "HIRES"


class ResultStatus:
    """
    Result status
    """

    def __init__(self, status: StatusCode, path: Path) -> None:
        self.status = status
        self.path = path

    def __repr__(self) -> str:
        return f"{self.status} : {self.path.name}"

    def print(self) -> None:
        """Print repr (for rich package)"""
        print(self.status, ": ", self.path.name)



"""
Absfuyu: Package data
---------------------
Load package data

Version: 5.17.0
Date updated: 20/11/2025 (dd/mm/yyyy)
"""

# Module level
# ---------------------------------------------------------------------------
__all__ = [
    "Pickler",
    "BasicLZMAOperation",
    "DataLoader",
    # "DataList",  # Disable due to Sphinx show real path, still importable
]


# Library
# ---------------------------------------------------------------------------
import lzma
import pickle
from enum import Enum
from importlib.resources import files
from pathlib import Path
from typing import Any

from absfuyu.core import BaseClass, versionadded


# Class
# ---------------------------------------------------------------------------
class DataList(Enum):
    CHEMISTRY = files("absfuyu.pkg_data").joinpath("chemistry.pkl")
    TAROT = files("absfuyu.pkg_data").joinpath("tarot.pkl")
    PASSWORDLIB = files("absfuyu.pkg_data").joinpath("passwordlib_lzma.pkl")


class Pickler(BaseClass):
    """
    A utility class for saving and loading data using the pickle format.

    This class provides static methods for serializing Python objects to a file
    using the ``pickle`` module and deserializing them back from a file.
    It simplifies the process of saving and loading data to and from pickle files.
    """

    @staticmethod
    def save(location: Path, data: Any) -> None:
        """
        Serializes and saves the given data to a file using the pickle format.

        Parameters
        ----------
        location : Path
            The path to the file where the data will be saved.

        data : Any
            The Python object to be serialized and saved.
        """
        with open(Path(location), "wb") as file:
            pickle.dump(data, file)

    @staticmethod
    def load(location: Path):
        """
        Loads and deserializes data from a file that is in pickle format.

        Parameters
        ----------
        location : Path
            The path to the pickle file to load.

        Returns
        -------
        Any
            The deserialized Python object loaded from the file.
        """
        with open(Path(location), "rb") as file:
            data = pickle.load(file)
        return data


@versionadded("5.0.0")
class BasicLZMAOperation(BaseClass):
    """
    A class for basic LZMA compression and decompression operations,
    integrated with pickle for saving and loading compressed data.

    This class provides static methods to compress data using LZMA,
    save the compressed data to a pickle file, load compressed data
    from a pickle file, and decompress LZMA-compressed data.
    """

    @staticmethod
    def save_to_pickle(location: Path, data: bytes) -> None:
        """
        Compresses the given byte data using LZMA and saves the compressed
        data to a file using pickle serialization.

        Parameters
        ----------
        location : Path
            The path to the file where the compressed data will be saved.

        data : bytes
            The byte data to be compressed and saved.


        Example:
        --------
        >>> data = b"This is some example data to compress."
        >>> BasicLZMAOperation.save_to_pickle(Path("compressed_data.pkl"), data)
        """
        compressed_data = lzma.compress(data)
        Pickler.save(location=location, data=compressed_data)

    @staticmethod
    def load(data: Any) -> bytes:
        """
        Decompresses LZMA-compressed data.

        Parameters
        ----------
        data : Any
            The LZMA-compressed data to be decompressed.

        Returns
        -------
        bytes
            The decompressed data.
        """
        return lzma.decompress(data)

    @staticmethod
    def load_from_pickle(location: Path) -> bytes:
        """
        Loads LZMA-compressed data from a pickle file, decompresses it.

        Parameters
        ----------
        location : Path
            The path to the pickle file containing the LZMA-compressed data.

        Returns
        -------
        bytes
            The decompressed data.
        """
        compressed_data = Pickler.load(location=location)
        return lzma.decompress(compressed_data)


@versionadded("5.0.0")
class DataLoader(BaseClass):
    """A class to load package data"""

    def __init__(self, data: DataList) -> None:
        """
        Initializes the DataLoader with a DataList object.

        Parameters
        ----------
        data : DataList
            A ``DataList`` object containing information about the data file to be loaded.
            It is expected to have a ``'value'`` attribute representing the file path.
        """
        self.data = data

    def load(self):
        """
        Loads data from the specified file, handling different file formats automatically.

        Returns
        -------
        Any
            The loaded data from the file.
        """
        try:
            data_name: str = self.data.value.stem  # type: ignore
        except Exception:
            data_name: str = self.data.value.name.split(".")[0]  # type: ignore

        if data_name.endswith("lzma"):
            return BasicLZMAOperation.load_from_pickle(self.data.value)  # type: ignore
        return Pickler.load(self.data.value)  # type: ignore



"""
Test: Config

Version: 5.17.0
Date updated: 20/11/2025 (dd/mm/yyyy)
"""

import random

import pytest

from absfuyu.config import ABSFUYU_CONFIG, Setting


# Setting
def test_from_dict():
    _ = Setting.from_dict({"test": {"default": False, "help": "HELP", "value": True}})
    assert True


# Config
class TestConfig:
    def test_add_and_del_setting(self) -> None:
        # Make random name
        name = f"test_add_setting_{random.randint(100_000, 999_999)}"

        # Test add
        old = len(ABSFUYU_CONFIG.settings)
        ABSFUYU_CONFIG.add_setting(name, True, True)
        new = len(ABSFUYU_CONFIG.settings)
        add_result = old < new

        # Test del
        ABSFUYU_CONFIG.del_setting(name)
        new2 = len(ABSFUYU_CONFIG.settings)
        del_result = old == new2

        # Output
        assert all([add_result, del_result])

    def test_toggle_setting(self) -> None:
        setting = "test"
        test_before = ABSFUYU_CONFIG._get_setting(setting).value
        ABSFUYU_CONFIG.toggle_setting(setting)
        test_after = ABSFUYU_CONFIG._get_setting(setting).value
        ABSFUYU_CONFIG.toggle_setting(setting)  # Back to original value
        assert test_before != test_after

    def test_reset_config(self) -> None:
        ABSFUYU_CONFIG.reset_config()
        test = [setting.value == setting.default for setting in ABSFUYU_CONFIG.settings]
        assert all(test)



"""
Absfuyu: Sort
-------------
Sort Module

Version: 5.17.0
Date updated: 20/11/2025 (dd/mm/yyyy)
"""

# Module level
# ---------------------------------------------------------------------------
__all__ = [
    # Sort
    "selection_sort",
    "insertion_sort",
    # Search
    "linear_search",
    "binary_search",
]


# Library
# ---------------------------------------------------------------------------
from typing import Any


# Functions
# ---------------------------------------------------------------------------
def selection_sort(iterable: list, reverse: bool = False) -> list:
    """
    Sort the list with selection sort (bubble sort) algorithm

    Parameters
    ----------
    iterable : list
        List that want to be sorted

    reverse : bool
        | if ``True``: sort in descending order
        | if ``False``: sort in ascending order
        | (default: ``False``)

    Returns
    -------
    list
        sorted list
    """

    if reverse:  # descending order
        for i in range(len(iterable)):
            for j in range(i + 1, len(iterable)):
                if iterable[i] < iterable[j]:
                    iterable[i], iterable[j] = iterable[j], iterable[i]
        return iterable

    else:  # ascending order
        for i in range(len(iterable)):
            for j in range(i + 1, len(iterable)):
                if iterable[i] > iterable[j]:
                    iterable[i], iterable[j] = iterable[j], iterable[i]
        return iterable


def insertion_sort(iterable: list) -> list:
    """
    Sort the list with insertion sort algorithm

    Parameters
    ----------
    iterable : list
        List that want to be sorted

    Returns
    -------
    list
        sorted list (ascending order)
    """

    for i in range(1, len(iterable)):
        key = iterable[i]
        j = i - 1
        while j >= 0 and key < iterable[j]:
            iterable[j + 1] = iterable[j]
            j -= 1
        iterable[j + 1] = key
    return iterable


def linear_search(iterable: list, key: Any) -> int:
    """
    Returns the position of ``key`` in the list

    Parameters
    ----------
    iterable : list
        List want to search

    key: Any
        Item want to find

    Returns
    -------
    int
        The position of ``key`` in the list if found, ``-1`` otherwise
    """
    for i, item in enumerate(iterable):
        if item == key:
            return i
    return -1


def binary_search(iterable: list, key: Any) -> int:
    """
    Returns the position of ``key`` in the list (list must be sorted)

    Parameters
    ----------
    iterable : list
        List want to search

    key: Any
        Item want to find

    Returns
    -------
    int
        The position of ``key`` in the list if found, ``-1`` otherwise
    """
    left = 0
    right = len(iterable) - 1
    while left <= right:
        middle = (left + right) // 2

        if iterable[middle] == key:
            return middle
        if iterable[middle] > key:
            right = middle - 1
        if iterable[middle] < key:
            left = middle + 1
    return -1



"""
Absfuyu: Core
-------------
Dummy cli

Version: 5.17.0
Date updated: 20/11/2025 (dd/mm/yyyy)
"""

# Module Package
# ---------------------------------------------------------------------------
__all__ = ["cli"]


# Library
# ---------------------------------------------------------------------------
import sys
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser

from absfuyu import __title__, __version__


# Function
# ---------------------------------------------------------------------------
def get_parser(
    name: str | None = None,
    description: str | None = None,
    epilog: str | None = None,
    *,
    version: str = "",
    add_help: bool = True,
    add_logging: bool = True,
) -> ArgumentParser:
    arg_parser = ArgumentParser(
        prog=name,
        description=description,
        epilog=epilog,
        add_help=add_help,
        formatter_class=ArgumentDefaultsHelpFormatter,
        # allow_abbrev=False, # Disable long options recognize
        # exit_on_error=True
    )
    arg_parser.add_argument(
        "--version", action="version", version=f"%(prog)s {version}"
    )
    if add_logging:
        _ll_val = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
        arg_parser.add_argument(
            "--log-level",
            metavar="LOG_LEVEL",
            dest="log_level",
            choices=_ll_val,
            default="INFO",
            help=f"Log level: {_ll_val}",
        )
    return arg_parser


def cli() -> None:
    desc = "This is a dummy cli, install <click> and <colorama> package to use this feature"
    arg_parser = get_parser(
        name=__title__,
        description=desc,
        version=__version__,
        add_logging=False,
    )
    args = arg_parser.parse_args(args=None if sys.argv[1:] else ["--help"])  # noqa



"""
Test: Everything

Version: 5.17.0
Date updated: 20/11/2025 (dd/mm/yyyy)
"""

# Library
# ---------------------------------------------------------------------------
import pytest

try:  # [beautiful] feature
    import rich  # type: ignore
except ImportError:
    rich = pytest.importorskip("rich")


# --- Loading test --------------------------------------------------------
from absfuyu import __author__, __license__, __title__, __version__
from absfuyu.config import (
    _SPACE_REPLACE,
    ABSFUYU_CONFIG,
    Config,
    ConfigFormat,
    Setting,
    SettingDictFormat,
)
from absfuyu.core import (
    BaseClass,
    CLITextColor,
    GetClassMembersMixin,
    __package_feature__,
    deprecated,
    tqdm,
    unidecode,
    versionadded,
    versionchanged,
)
from absfuyu.core.baseclass import (
    AutoREPRMixin,
    BaseDataclass,
    ClassMembers,
    ClassMembersResult,
    PositiveInitArgsMeta,
)
from absfuyu.core.baseclass2 import (
    PerformanceTrackingMeta,
    ShowAllMethodsMixinInspectVer,
    positive_class_init_args,
)
from absfuyu.core.decorator import (
    add_subclass_methods_decorator,
    dummy_decorator,
    dummy_decorator_with_args,
)
from absfuyu.core.docstring import (
    _SPHINX_DOCS_TEMPLATE,
    SphinxDocstring,
    SphinxDocstringMode,
)
from absfuyu.core.dummy_func import dummy_function
from absfuyu.dxt import (
    DictAnalyzeResult,
    DictBoolFalse,
    DictBoolTrue,
    DictExt,
    IntExt,
    ListExt,
    ListNoDunder,
    ListREPR,
    Text,
    TextAnalyzeDictFormat,
)
from absfuyu.extra import is_loaded
from absfuyu.extra.beautiful import BeautifulOutput  # Has rich
from absfuyu.extra.da.dadf import (  # Has pandas, numpy
    DADF,
    DataAnalystDataFrameCityMixin,
    DataAnalystDataFrameColumnMethodMixin,
    DataAnalystDataFrameDateMixin,
    DataAnalystDataFrameInfoMixin,
    DataAnalystDataFrameNAMixin,
    DataAnalystDataFrameOtherMixin,
    DataAnalystDataFrameRowMethodMixin,
)
from absfuyu.extra.da.dadf_base import (  # Has pandas, numpy
    CityData,
    DataAnalystDataFrameBase,
    SplittedDF,
)
from absfuyu.extra.da.mplt import MatplotlibFormatString, _PLTFormatString
from absfuyu.extra.data_analysis import (  # Has pandas, numpy
    compare_2_list,
    equalize_df,
    rename_with_dict,
)
from absfuyu.fun import happy_new_year, human_year_to_dog_year, zodiac_sign
from absfuyu.fun.rubik import (
    OLL,
    PLL,
    PLL_PIC,
    Cross,
    PLLs,
    Rubik3x3,
    RubikAlgorithm,
    RubikNotation,
    RubikNotations,
)
from absfuyu.fun.tarot import Tarot, TarotCard
from absfuyu.game import GameStats, game_escapeLoop, game_RockPaperScissors
from absfuyu.game.sudoku import Sudoku
from absfuyu.game.tictactoe import GameMode, TicTacToe
from absfuyu.game.wordle import Wordle  # Has requests

# --- Sub-package ---
from absfuyu.general.content import (  # Has unidecode
    Content,
    ContentLoader,
    LoadedContent,
)
from absfuyu.general.human import BloodType, Human, Person
from absfuyu.general.shape import (
    Circle,
    Cube,
    Cuboid,
    Cylinder,
    EqualSidesPolygon,
    HemiSphere,
    Hexagon,
    Parallelogram,
    Pentagon,
    Polygon,
    Rectangle,
    Rhombus,
    Shape,
    Sphere,
    Square,
    ThreeDimensionShape,
    Trapezoid,
    Triangle,
)

# from absfuyu.logger import *
from absfuyu.logger import LogLevel, compress_for_log, logger
from absfuyu.pkg_data import BasicLZMAOperation, DataList, DataLoader, Pickler
from absfuyu.sort import binary_search, insertion_sort, linear_search, selection_sort

# from absfuyu.tools import *
from absfuyu.tools.checksum import Checksum, ChecksumMode
from absfuyu.tools.converter import (
    Base64EncodeDecode,
    ChemistryElement,
    Str2Pixel,
    Text2Chemistry,
)
from absfuyu.tools.generator import Charset, Generator
from absfuyu.tools.inspector import Inspector, inspect_all
from absfuyu.tools.keygen import Keygen
from absfuyu.tools.obfuscator import Obfuscator, StrShifter
from absfuyu.tools.passwordlib import TOTP, PasswordGenerator, PasswordHash
from absfuyu.tools.shutdownizer import (
    ShutdownEngine,
    ShutdownEngineLinux,
    ShutdownEngineMac,
    ShutdownEngineWin,
    ShutDownizer,
)
from absfuyu.tools.web import gen_random_commit_msg, soup_link  # Has bs4, requests
from absfuyu.typings import (
    _CALLABLE,
    CT,
    KT,
    VT,
    N,
    P,
    R,
    SupportsShowMethods,
    T,
    T_co,
    T_contra,
    _Number,
)
from absfuyu.util import (
    convert_to_raw_unicode,
    get_installed_package,
    set_min_max,
    stop_after_day,
)
from absfuyu.util.api import APIRequest, ping_windows  # Has requests
from absfuyu.util.json_method import JsonFile
from absfuyu.util.lunar import LunarCalendar
from absfuyu.util.path import (
    Directory,
    DirectoryArchiverMixin,
    DirectoryBase,
    DirectoryBasicOperationMixin,
    DirectoryInfo,
    DirectoryInfoMixin,
    DirectoryOrganizerMixin,
    DirectoryTreeMixin,
    FileOrFolderWithModificationTime,
    SaveFileAs,
)
from absfuyu.util.performance import (
    Checker,
    function_benchmark,
    function_debug,
    measure_performance,
    retry,
)
from absfuyu.util.shorten_number import (
    CommonUnitSuffixesFactory,
    Decimal,
    UnitSuffixFactory,
    shorten_number,
)
from absfuyu.util.text_table import (
    BoxDrawingCharacterBase,
    BoxDrawingCharacterBold,
    BoxDrawingCharacterDashed,
    BoxDrawingCharacterDashed3,
    BoxDrawingCharacterDashed4,
    BoxDrawingCharacterDashedBold,
    BoxDrawingCharacterDashedRound,
    BoxDrawingCharacterDiamond,
    BoxDrawingCharacterDouble,
    BoxDrawingCharacterNormal,
    BoxDrawingCharacterRounded,
    BoxStyle,
    OneColumnTableMaker,
    get_box_drawing_character,
)
from absfuyu.version import (
    Bumper,
    PkgVersion,
    ReleaseLevel,
    ReleaseOption,
    Version,
    VersionDictFormat,
)


# Test
# ---------------------------------------------------------------------------
def test_everything():
    assert True



"""
ABSFUYU CLI
-----------
Color

Version: 5.17.0
Date updated: 20/11/2025 (dd/mm/yyyy)
"""

# Module Package
# ---------------------------------------------------------------------------
__all__ = ["COLOR"]


# Library
# ---------------------------------------------------------------------------
try:
    import colorama
except ImportError:  # Check for `colorama`
    colorama = None

# Color
# ---------------------------------------------------------------------------
if colorama is None:
    COLOR = {
        "green": "",
        "GREEN": "",
        "blue": "",
        "BLUE": "",
        "red": "",
        "RED": "",
        "yellow": "",
        "YELLOW": "",
        "reset": "",
    }
else:
    COLOR = {
        "green": colorama.Fore.LIGHTGREEN_EX,
        "GREEN": colorama.Fore.GREEN,
        "blue": colorama.Fore.LIGHTCYAN_EX,
        "BLUE": colorama.Fore.CYAN,
        "red": colorama.Fore.LIGHTRED_EX,
        "RED": colorama.Fore.RED,
        "yellow": colorama.Fore.LIGHTYELLOW_EX,
        "YELLOW": colorama.Fore.YELLOW,
        "reset": colorama.Fore.RESET,
    }



"""
Test: Util
----------
Text table

Version: 5.17.0
Date updated: 20/11/2025 (dd/mm/yyyy)
"""

import os

import pytest

from absfuyu.util.text_table import BoxDrawingCharacterNormal, OneColumnTableMaker


@pytest.fixture
def table_maker():
    return OneColumnTableMaker(ncols=None, style="normal")


@pytest.mark.abs_util
class TestOneColumnTableMaker:
    """absfuyu.util.text_table.OneColumnTableMaker"""

    def test_init_default_ncols(self, table_maker: OneColumnTableMaker) -> None:
        assert table_maker.ncols >= 5

    def test_init_custom_ncols(self) -> None:
        maker = OneColumnTableMaker(ncols=10)
        assert maker.ncols == 10

    def test_init_min_ncols(self) -> None:
        maker = OneColumnTableMaker(ncols=3)
        assert maker.ncols == 5

    def test_add_title_short(self, table_maker: OneColumnTableMaker) -> None:
        table_maker.add_title("Short Title")
        assert len(table_maker._title) == table_maker.ncols

    def test_add_title_long(self, table_maker: OneColumnTableMaker) -> None:
        mul = table_maker.ncols + 5
        table_maker.add_title("A" * mul)
        assert len(table_maker._title) == table_maker.ncols

    def test_add_paragraph_single_string(
        self, table_maker: OneColumnTableMaker
    ) -> None:
        table_maker.add_paragraph("Hello")
        assert table_maker._paragraphs == [["Hello"]]

    def test_add_paragraph_multiple_strings(
        self, table_maker: OneColumnTableMaker
    ) -> None:
        table_maker.add_paragraph(["Hello", "World"])
        assert table_maker._paragraphs == [["Hello", "World"]]

    def test_make_line_upper(self, table_maker: OneColumnTableMaker) -> None:
        line = table_maker._make_line(0)
        assert line.startswith(BoxDrawingCharacterNormal.UPPER_LEFT_CORNER)
        assert line.endswith(BoxDrawingCharacterNormal.UPPER_RIGHT_CORNER)

    def test_make_line_intersect(self, table_maker: OneColumnTableMaker) -> None:
        line = table_maker._make_line(1)
        assert line.startswith(BoxDrawingCharacterNormal.VERTICAL_RIGHT)
        assert line.endswith(BoxDrawingCharacterNormal.VERTICAL_LEFT)

    def test_make_line_lower(self, table_maker: OneColumnTableMaker) -> None:
        line = table_maker._make_line(2)
        assert line.startswith(BoxDrawingCharacterNormal.LOWER_LEFT_CORNER)
        assert line.endswith(BoxDrawingCharacterNormal.LOWER_RIGHT_CORNER)

    def test_make_table_empty(self, table_maker: OneColumnTableMaker) -> None:
        assert table_maker._make_table() is None

    def test_make_table_single_paragraph(
        self, table_maker: OneColumnTableMaker
    ) -> None:
        table_maker.add_paragraph(["Hello", "World"])
        table = table_maker._make_table()
        assert table is not None
        assert len(table) > 0

    def test_make_table_multiple_paragraphs(
        self, table_maker: OneColumnTableMaker
    ) -> None:
        table_maker.add_paragraph(["Hello", "World"])
        table_maker.add_paragraph(["Foo", "Bar"])
        table = table_maker._make_table()
        assert table is not None
        assert len(table) > 0

    def test_make_table_with_title(self, table_maker: OneColumnTableMaker) -> None:
        table_maker.add_title("My Title")
        table_maker.add_paragraph(["Hello", "World"])
        table = table_maker._make_table()
        assert table is not None
        assert len(table) > 0

    def test_make_table_output(self, table_maker: OneColumnTableMaker) -> None:
        table_maker.add_paragraph(["Hello", "World"])
        output = table_maker.make_table()
        assert isinstance(output, str)
        assert len(output) > 0

    @pytest.mark.parametrize("terminal_size", [80, 120])
    def test_init_terminal_size_mock(self, mocker, terminal_size: int) -> None:
        mock_get_size = mocker.patch("os.get_terminal_size")
        # mock_get_size.side_effect = OSError("")
        mock_get_size.return_value.columns = terminal_size
        # with pytest.raises(OSError):
        #     assert maker.ncols == 88
        maker = OneColumnTableMaker()
        assert maker.ncols == terminal_size



"""
Absfuyu: Web
------------
Web, ``request``, ``BeautifulSoup`` stuff

Version: 5.17.0
Date updated: 20/11/2025 (dd/mm/yyyy)
"""

# Library
# ---------------------------------------------------------------------------
import requests
from bs4 import BeautifulSoup

from absfuyu.logger import logger


# Function
# ---------------------------------------------------------------------------
def soup_link(link: str) -> BeautifulSoup:
    """
    ``BeautifulSoup`` the link

    Parameters
    ----------
    link : str
        Link to BeautifulSoup

    Returns
    -------
    BeautifulSoup
        ``BeautifulSoup`` instance
    """
    try:
        page = requests.get(link)
        soup = BeautifulSoup(page.content, "html.parser")
        logger.debug("Soup completed!")
        return soup
    except Exception:
        logger.error("Can't soup")
        raise SystemExit("Something wrong")  # noqa: B904


def gen_random_commit_msg() -> str:
    """
    Generate random commit message

    Returns
    -------
    str
        Random commit message
    """
    out = soup_link("https://whatthecommit.com/").get_text()[34:-20]
    logger.debug(out)
    return out  # type: ignore



"""
Absfuyu: Data Extension
-----------------------
Support classes

Version: 5.17.0
Date updated: 20/11/2025 (dd/mm/yyyy)
"""

# Module Package
# ---------------------------------------------------------------------------
__all__ = ["DictBoolTrue", "DictBoolFalse"]


# Library
# ---------------------------------------------------------------------------
from typing import Any

from absfuyu.logger import _compress_list_for_print


# Dict Boolean Masking Repr
# ---------------------------------------------------------------------------
def _dict_bool(dict_object: dict, option: bool) -> dict | None:
    """
    Support function DictBool class
    """
    out = dict()
    for k, v in dict_object.items():
        if v == option:
            out[k] = v
    if out:
        return out
    else:
        return None


class DictBoolTrue(dict[Any, bool]):
    """Only show items when ``values == True`` in ``__repr__()``"""

    def __repr__(self) -> str:
        temp = self.copy()
        return _dict_bool(temp, True).__repr__()


class DictBoolFalse(dict[Any, bool]):
    """Only show items when ``values == False`` in ``__repr__()``"""

    def __repr__(self) -> str:
        temp = self.copy()
        return _dict_bool(temp, False).__repr__()


# ---------------------------------------------------------------------------
class ListREPR(list):
    """Show ``list`` in shorter form"""

    def __repr__(self) -> str:
        return _compress_list_for_print(self, 9)


class ListNoDunder(list[str]):
    """Use with ``object.__dir__()``"""

    def __repr__(self) -> str:
        out = [x for x in self if not x.startswith("__")]
        return out.__repr__()


# class DictNoDunder(dict):  # W.I.P
#     """Remove dunder methods in ``__repr__()`` of dict"""

#     def __repr__(self) -> str:
#             temp = self.copy()
#             out = dict()
#             for k, v in temp.items():
#                 if not str(k).startswith("__"):
#                     out.__setattr__(k, v)
#             return out.__repr__()



"""
Absfuyu: Audio
--------------
Audio lossless checker

Version: 5.17.0
Date updated: 22/11/2025 (dd/mm/yyyy)
"""

# Module level
# ---------------------------------------------------------------------------
__all__ = ["AudioInfo", "DirectoryAudioLosslessCheckMixin"]


# Library
# ---------------------------------------------------------------------------
import json
import subprocess
from concurrent.futures import ProcessPoolExecutor, as_completed
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, NamedTuple

from absfuyu.core.dummy_func import tqdm
from absfuyu.extra.audio._util import ResultStatus, StatusCode
from absfuyu.util import is_command_available
from absfuyu.util.path import DirectoryBase
from absfuyu.util.shorten_number import Duration

try:
    import numpy as np
except ImportError:
    from subprocess import run

    from absfuyu.config import ABSFUYU_CONFIG

    if ABSFUYU_CONFIG._get_setting("auto-install-extra").value:  # type: ignore
        cmd = "python -m pip install -U absfuyu[extra]".split()
        run(cmd)
    else:
        raise SystemExit("This feature is in absfuyu[extra] package")  # noqa: B904


# Class
# ---------------------------------------------------------------------------
class FrequencyRange(NamedTuple):
    """Audio frequency range"""

    min: int | float
    max: int | float


@dataclass
class AudioInfo:
    """
    Audio infomation

    Parameters
    ----------
    sample_rate : int | None, optional
        Sample rate (Hz)

    bit_depth : int, optional
        Bit depth

    channels : int | None, optional
        Number of channels

    codec : str | None, optional
        Audio codec

    duration_raw : float, optional
        Duration of audio (second)

    bitrate : Any | None, optional
        Bitrate

    audio_path : Path | None, optional
        Path to audio

    Returns
    -------
    AudioInfo
        Audio infomation
    """

    sample_rate: int | None = field(default=None, metadata={"unit": "Hz"})
    bit_depth: int = field(default=0, metadata={"unit": "bit"})
    channels: int | None = field(default=None)
    codec: str | None = field(default=None)
    duration_raw: float = field(default=0.0, repr=False, metadata={"unit": "second"})
    bitrate: int | None = field(default=None)
    audio_path: Path | None = field(default=None, repr=False)  # hide path in repr

    # computed fields
    duration: float = field(init=False)
    freq_range: FrequencyRange = field(init=False, metadata={"unit": "Hz"})

    def __post_init__(self):
        self.duration = Duration(self.duration_raw)

        # Resource intensive method
        # _, sampling_rate = librosa.load(audio_path, sr=None)
        # freqs = librosa.fft_frequencies(sr=sampling_rate)
        freqs = np.fft.rfftfreq(n=2048, d=1.0 / self.sample_rate)
        self.freq_range = FrequencyRange(freqs[0], freqs[-1])

    @property
    def is_lossless(self) -> bool:
        """
        Audio is lossless when frequencies above 20,000 are not cut off

        Returns
        -------
        bool
            If audio is lossless
        """
        return self.freq_range[1] >= 20000

    @property
    def is_hi_res(self) -> bool:
        """
        Audio is HiRes when lossless and have sample rate >= 48,000Hz or bit rate >= 16 bit

        Returns
        -------
        bool
            If audio is lossless
        """
        bd = 0 if self.bit_depth is None else self.bit_depth
        return all([self.sample_rate >= 48000, bd >= 24, self.is_lossless])


class DirectoryAudioLosslessCheckMixin(DirectoryBase):
    """
    Directory - Audio lossless checker

    - lossless_check
    """

    def get_audio_info(self, audio_path: Path, /) -> AudioInfo:
        """
        Return audio info using ffprobe.

        Parameters
        ----------
        audio_path : Path
            Path to audio

        Returns
        -------
        AudioInfo
            Audio infomation
        """

        is_command_available(["ffmpeg"], "ERROR: ffmpeg not installed to PATH")

        # Probe
        cmd = [
            "ffprobe",
            "-v",
            "error",
            "-print_format",
            "json",
            "-show_streams",
            "-select_streams",
            "a:0",
            str(audio_path.resolve()),
        ]

        result = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
        data: dict[str, Any] = json.loads(result.stdout)["streams"][0]

        bit_depth = data.get("bits_per_sample") or data.get("bits_per_raw_sample")
        if bit_depth is not None:
            bit_depth = int(bit_depth)

        ai = AudioInfo(
            sample_rate=int(data.get("sample_rate", 0)),
            bit_depth=bit_depth,
            channels=data.get("channels"),
            codec=data.get("codec_name"),
            duration_raw=float(data["duration"]) if "duration" in data else None,
            bitrate=int(data["bit_rate"]) if "bit_rate" in data else None,
            audio_path=audio_path,
        )
        return ai

    def lossless_check_one(self, audio_path: Path) -> ResultStatus:
        """
        Check if audio is lossless

        Parameters
        ----------
        audio_path : Path
            Path to audio

        Returns
        -------
        ResultStatus
            If audio is lossless
        """
        res = self.get_audio_info(audio_path)
        if res.is_hi_res:
            return ResultStatus(StatusCode.HIRES, audio_path)
        elif res.is_lossless:
            return ResultStatus(StatusCode.LOSSLESS, audio_path)
        return ResultStatus(StatusCode.NOT_LOSSLESS, audio_path)

    def lossless_check_single_thread(self, from_format: str = ".flac", recursive: bool = True) -> None:
        """
        Check if audios in directory are lossless - single thread version

        Parameters
        ----------
        from_format : str, optional
            Audio format, by default ".flac"

        recursive : bool, optional
            Include audio in child folder, by default True
        """
        audios = list(self.source_path.rglob(f"{'**/*'if recursive else '*'}{from_format}"))

        if not audios:
            print(f"No {from_format.upper()} files found.")
            return None

        print(f"Found {len(audios)} {from_format.upper()} files.")

        results: list[ResultStatus] = []
        for x in tqdm(audios, desc="Checking", unit="file", unit_scale=True):
            results.append(self.lossless_check_one(x))

        print("\n--- Summary ---")
        for line in results:
            line.print()

    def lossless_check(
        self,
        from_format: str = ".flac",
        recursive: bool = True,
        workers: int | None = None,
    ) -> None:
        """
        Check if audios in directory are lossless

        Parameters
        ----------
        from_format : str, optional
            Audio format, by default ".flac"

        recursive : bool, optional
            Include audio in child folder, by default True

        workers : int | None, optional
            Number of parallel processing threads, by default None
        """
        audios = list(self.source_path.rglob(f"{'**/*'if recursive else '*'}{from_format}"))

        if not audios:
            print(f"No {from_format.upper()} files found.")
            return None

        print(f"Found {len(audios)} {from_format.upper()} files.")

        results: list[ResultStatus] = []
        with ProcessPoolExecutor(max_workers=workers) as executor:
            futures = {executor.submit(self.lossless_check_one, x): x for x in audios}

            for fut in tqdm(
                as_completed(futures),
                total=len(futures),
                desc="Checking",
                unit="file",
                unit_scale=True,
            ):
                results.append(fut.result())

        print("\n--- Summary ---")
        for line in results:
            line.print()



"""
Absufyu: Utilities
------------------
Text table

Version: 5.17.0
Date updated: 20/11/2025 (dd/mm/yyyy)
"""

# Module level
# ---------------------------------------------------------------------------
__all__ = ["BoxStyle", "OneColumnTableMaker"]


# Library
# ---------------------------------------------------------------------------
import os
from collections.abc import Sequence
from textwrap import TextWrapper
from typing import Literal

from absfuyu.core import BaseClass

# Style
# ---------------------------------------------------------------------------
BoxStyle = Literal[
    "normal",
    "bold",
    "dashed",
    "double",
    "rounded",
    "drounded",
    "diamond",
    "dbold",
    "dashed3",
    "dashed4",
]


class BoxDrawingCharacterBase:
    """
    Box drawing characters - Base/Normal characters

    Characters reference: https://en.wikipedia.org/wiki/Box-drawing_characters
    """

    UPPER_LEFT_CORNER = "\u250c"
    UPPER_RIGHT_CORNER = "\u2510"
    HORIZONTAL = "\u2500"
    VERTICAL = "\u2502"
    LOWER_LEFT_CORNER = "\u2514"
    LOWER_RIGHT_CORNER = "\u2518"
    VERTICAL_RIGHT = "\u251c"
    VERTICAL_LEFT = "\u2524"
    CROSS = "\u253c"
    HORIZONTAL_UP = "\u2534"
    HORIZONTAL_DOWN = "\u252c"


class BoxDrawingCharacterNormal(BoxDrawingCharacterBase):
    """Normal"""

    pass


class BoxDrawingCharacterDashed(BoxDrawingCharacterNormal):
    """Dashed"""

    HORIZONTAL = "\u254c"
    VERTICAL = "\u254e"


class BoxDrawingCharacterDashed3(BoxDrawingCharacterNormal):
    """Triple dashed"""

    HORIZONTAL = "\u2504"
    VERTICAL = "\u2506"


class BoxDrawingCharacterDashed4(BoxDrawingCharacterNormal):
    """Quadruple dashed"""

    HORIZONTAL = "\u2508"
    VERTICAL = "\u250a"


class BoxDrawingCharacterRounded(BoxDrawingCharacterNormal):
    """Rounded"""

    UPPER_LEFT_CORNER = "\u256d"
    UPPER_RIGHT_CORNER = "\u256e"
    LOWER_LEFT_CORNER = "\u2570"
    LOWER_RIGHT_CORNER = "\u256f"


class BoxDrawingCharacterDiamond(BoxDrawingCharacterNormal):
    """Diamond"""

    UPPER_LEFT_CORNER = "\u2571"
    UPPER_RIGHT_CORNER = "\u2572"
    LOWER_LEFT_CORNER = "\u2572"
    LOWER_RIGHT_CORNER = "\u2571"


class BoxDrawingCharacterDashedRound(
    BoxDrawingCharacterDashed, BoxDrawingCharacterRounded
):
    """Dashed rounded"""

    pass


class BoxDrawingCharacterBold(BoxDrawingCharacterBase):
    """Bold"""

    UPPER_LEFT_CORNER = "\u250f"
    UPPER_RIGHT_CORNER = "\u2513"
    HORIZONTAL = "\u2501"
    VERTICAL = "\u2503"
    LOWER_LEFT_CORNER = "\u2517"
    LOWER_RIGHT_CORNER = "\u251b"
    VERTICAL_RIGHT = "\u2523"
    VERTICAL_LEFT = "\u252b"
    CROSS = "\u254b"
    HORIZONTAL_UP = "\u253b"
    HORIZONTAL_DOWN = "\u2533"


class BoxDrawingCharacterDashedBold(BoxDrawingCharacterBold):
    """Dashed bold"""

    HORIZONTAL = "\u254d"
    VERTICAL = "\u254f"


class BoxDrawingCharacterDouble(BoxDrawingCharacterBase):
    """Double"""

    UPPER_LEFT_CORNER = "\u2554"
    UPPER_RIGHT_CORNER = "\u2557"
    HORIZONTAL = "\u2550"
    VERTICAL = "\u2551"
    LOWER_LEFT_CORNER = "\u255a"
    LOWER_RIGHT_CORNER = "\u255d"
    VERTICAL_RIGHT = "\u2560"
    VERTICAL_LEFT = "\u2563"
    CROSS = "\u256c"
    HORIZONTAL_UP = "\u2569"
    HORIZONTAL_DOWN = "\u2566"


def get_box_drawing_character(
    style: BoxStyle | str = "normal",
) -> type[BoxDrawingCharacterBase]:
    """
    Choose style for Box drawing characters.

    Parameters
    ----------
    style : BoxStyle | str, optional
        Style for the table, by default ``"normal"``.
        Available style:
        - normal
        - bold
        - dashed
        - double
        - rounded
        - drounded: Dashed and rounded
        - diamond
        - dbold: Bold dashed
        - dashed3: Triple dash
        - dashed4: Quadruple dash

    Returns
    -------
    BoxDrawingCharacter
        Box drawing characters in specified style.
    """

    match style.lower().strip():
        case "normal":
            return BoxDrawingCharacterNormal
        case "bold":
            return BoxDrawingCharacterBold
        case "dashed":
            return BoxDrawingCharacterDashed
        case "dashed3":
            return BoxDrawingCharacterDashed3
        case "dashed4":
            return BoxDrawingCharacterDashed4
        case "double":
            return BoxDrawingCharacterDouble
        case "rounded":
            return BoxDrawingCharacterRounded
        case "drounded":
            return BoxDrawingCharacterDashedRound
        case "diamond":
            return BoxDrawingCharacterDiamond
        case "dbold":
            return BoxDrawingCharacterDashedBold
        case _:
            return BoxDrawingCharacterNormal


# Class
# ---------------------------------------------------------------------------
class OneColumnTableMaker(BaseClass):
    """
    Table Maker instance

    Parameters
    ----------
    ncols : int | None, optional
        Length of the table (include content). Must be >= 5.
        Set to ``None`` to use maximum length,
        defaults to ``88`` when failed to use ``os.get_terminal_size()``.
        By default ``None``

    BoxStyle : Literal["normal", "bold", "dashed", "double"], optional
        Style for the table, by default ``"normal"``
    """

    __slots__ = ("ncols", "_title", "_paragraphs", "_table_char", "_text_wrapper")

    def __init__(self, ncols: int | None = None, style: BoxStyle = "normal") -> None:
        """
        Table Maker instance

        Parameters
        ----------
        ncols : int | None, optional
            Length of the table (include content). Must be >= 5.
            Set to ``None`` to use maximum length,
            defaults to ``88`` when failed to use ``os.get_terminal_size()``.
            By default ``None``

        style : BoxStyle, optional
            Style for the table, by default ``"normal"``
        """

        # Text length
        if ncols is None:
            try:
                self.ncols = os.get_terminal_size().columns
            except OSError:
                self.ncols = 88
        else:
            self.ncols = max(5, ncols)

        # Title & paragraph
        self._title = ""
        self._paragraphs: list[Sequence[str]] = []

        # Style
        self._table_char = get_box_drawing_character(style=style)

        # Text wrapper
        self._text_wrapper = TextWrapper(
            width=self.ncols - 4,
            initial_indent="",
            subsequent_indent="",
            tabsize=4,
            break_long_words=True,
        )

    def add_title(self, title: str) -> None:
        """
        Add title to Table

        Parameters
        ----------
        title : str
            Title to add.
            When ``len(title) > ncols``: title will not show
        """
        max_padding_length = self.ncols - 2
        if max_padding_length < (len(title) + 2) or len(title) < 1:
            _title = ""
        else:
            _title = f" {title} "

        line = (
            f"{self._table_char.UPPER_LEFT_CORNER}"
            f"{_title.center(max_padding_length, self._table_char.HORIZONTAL)}"
            f"{self._table_char.UPPER_RIGHT_CORNER}"
        )
        self._title = line

    def add_paragraph(self, paragraph: Sequence[str]) -> None:
        """
        Add paragraph into Table

        Parameters
        ----------
        paragraph : Sequence[str]
            An iterable of str
        """
        if isinstance(paragraph, str):
            self._paragraphs.append([paragraph])
        else:
            self._paragraphs.append(paragraph)

    def _make_line(self, option: Literal[0, 1, 2]) -> str:
        options = (
            (self._table_char.UPPER_LEFT_CORNER, self._table_char.UPPER_RIGHT_CORNER),
            (self._table_char.VERTICAL_RIGHT, self._table_char.VERTICAL_LEFT),
            (self._table_char.LOWER_LEFT_CORNER, self._table_char.LOWER_RIGHT_CORNER),
        )
        max_line_length = self.ncols - 2
        line = (
            f"{options[option][0]}"
            f"{''.ljust(max_line_length, self._table_char.HORIZONTAL)}"
            f"{options[option][1]}"
        )
        return line

    def _make_table(self) -> list[str] | None:
        # Check if empty
        if len(self._paragraphs) < 1:
            return None
        if len(self._paragraphs[0]) < 1:
            return None

        # Make table
        max_content_length = self.ncols - 4
        paragraph_length = len(self._paragraphs)

        # Line prep
        _first_line = self._make_line(0)
        _sep_line = self._make_line(1)
        _last_line = self._make_line(2)

        # Table
        table: list[str] = [_first_line] if self._title == "" else [self._title]
        for i, paragraph in enumerate(self._paragraphs, start=1):
            for line in paragraph:
                splitted_line = self._text_wrapper.wrap(line) if len(line) > 0 else [""]
                mod_lines: list[str] = [
                    f"{self._table_char.VERTICAL} "
                    f"{line.ljust(max_content_length, ' ')}"
                    f" {self._table_char.VERTICAL}"
                    for line in splitted_line
                ]
                table.extend(mod_lines)

            if i != paragraph_length:
                table.append(_sep_line)
            else:
                table.append(_last_line)
        return table

    def make_table(self) -> str:
        table = self._make_table()
        if table is None:
            return ""
        return "\n".join(table)


# W.I.P
# ---------------------------------------------------------------------------
class _BoxDrawingCharacterFactory:
    _TRANSLATE: dict[str, str] = {
        "n": "normal",
        "d": "dashed",
        "b": "bold",
        "r": "rounded",
        "x": "double",
    }

    UPPER_LEFT_CORNER: list[tuple[str, str]] = [
        ("\u250c", "n,d"),
        ("\u256d", "r"),
        ("\u250f", "b"),
        ("\u2554", "x"),
    ]
    UPPER_RIGHT_CORNER: list[tuple[str, str]] = [
        ("\u2510", "n,d"),
        ("\u256e", "r"),
        ("\u2513", "b"),
        ("\u2557", "x"),
    ]
    HORIZONTAL: list[tuple[str, str]] = [
        ("\u2500", "n,r"),
        ("\u254c", "d"),
        ("\u2501", "b"),
        ("\u2550", "x"),
    ]
    VERTICAL: list[tuple[str, str]] = [
        ("\u2502", "n,r"),
        ("\u254e", "d"),
        ("\u2503", "b"),
        ("\u2551", "x"),
    ]
    LOWER_LEFT_CORNER: list[tuple[str, str]] = [
        ("\u2514", "n,d"),
        ("\u2570", "r"),
        ("\u2517", "b"),
        ("\u255a", "x"),
    ]
    LOWER_RIGHT_CORNER: list[tuple[str, str]] = [
        ("\u2518", "n,d"),
        ("\u256f", "r"),
        ("\u251b", "b"),
        ("\u255d", "x"),
    ]
    VERTICAL_RIGHT: list[tuple[str, str]] = [
        ("\u251c", "n,d,r"),
        ("\u2523", "b"),
        ("\u2560", "x"),
    ]
    VERTICAL_LEFT: list[tuple[str, str]] = [
        ("\u2524", "n,d,r"),
        ("\u252b", "b"),
        ("\u2563", "x"),
    ]
    CROSS: list[tuple[str, str]] = [
        ("\u253c", "n,d,r"),
        ("\u254b", "b"),
        ("\u256c", "x"),
    ]
    HORIZONTAL_UP: list[tuple[str, str]] = [
        ("\u2534", "n,d,r"),
        ("\u253b", "b"),
        ("\u2569", "x"),
    ]
    HORIZONTAL_DOWN: list[tuple[str, str]] = [
        ("\u252c", "n,d,r"),
        ("\u2533", "b"),
        ("\u2566", "x"),
    ]

    _FIELDS: tuple[str, ...] = (
        "UPPER_LEFT_CORNER",
        "UPPER_RIGHT_CORNER",
        "HORIZONTAL",
        "VERTICAL",
        "LOWER_LEFT_CORNER",
        "LOWER_RIGHT_CORNER",
        "VERTICAL_RIGHT",
        "VERTICAL_LEFT",
        "CROSS",
        "HORIZONTAL_UP",
        "HORIZONTAL_DOWN",
    )

    def __init__(self, style: str | None = None):
        self.style = style

    @classmethod
    def _make_style(cls) -> dict[str, dict[str, str]]:
        """
        Creates a style dictionary based on the class attributes.

        Returns
        -------
        dict[str, dict[str, str]]
            A dictionary mapping group names to style configurations.
        """

        # Initialize an empty style dictionary
        style: dict[str, dict[str, str]] = {}

        # Create a dictionary mapping field names to themselves
        field_map = {field: "" for field in cls._FIELDS}

        # Initialize style entries for each translation key
        for x in cls._TRANSLATE.keys():
            style[x] = field_map

        # Extract character data from class fields
        char_data: list[tuple[str, list[tuple[str, str]]]] = [
            (field, getattr(cls, field)) for field in cls._FIELDS
        ]

        # Populate the style dictionary with character mappings
        for name, chars in char_data:
            for char, groups in chars:
                for group in map(str.strip, groups.split(",")):
                    style[group][name] = char

        return style



"""
ABSFUYU CLI
-----------
Audio

Version: 5.17.0
Date updated: 22/11/2025 (dd/mm/yyyy)
"""

# Module Package
# ---------------------------------------------------------------------------
__all__ = ["audio_group"]


# Library
# ---------------------------------------------------------------------------
from pathlib import Path

import click

from absfuyu.extra.audio.convert import DirectoryAudioConvertMixin
from absfuyu.extra.audio.lossless import DirectoryAudioLosslessCheckMixin


# CLI
# ---------------------------------------------------------------------------
@click.command(name="c3")
@click.argument("dir_path", type=str)
@click.option(
    "--recursive",
    "-r",
    "recursive_mode",
    type=bool,
    default=False,
    is_flag=True,
    show_default=True,
    help="Scan for every file in the folder (including child folder)",
)
@click.option(
    "--format",
    "-f",
    "from_format",
    type=str,
    default=".flac",
    show_default=True,
    help="From which audio format",
)
def convert_to_mp3(dir_path: str, from_format: str, recursive_mode: bool) -> None:
    """Convert to .mp3 file"""

    engine = DirectoryAudioConvertMixin(Path(dir_path).resolve())

    try:
        engine.convert_to_mp3(from_format=from_format, recursive=recursive_mode)
    except Exception:
        engine.convert_to_mp3_single_thread(from_format=from_format, recursive=recursive_mode)
    click.echo("Done")


@click.command(name="lc")
@click.argument("dir_path", type=str)
@click.option(
    "--recursive",
    "-r",
    "recursive_mode",
    type=bool,
    default=False,
    is_flag=True,
    show_default=True,
    help="Scan for every file in the folder (including child folder)",
)
@click.option(
    "--format",
    "-f",
    "format",
    type=str,
    default=".flac",
    show_default=True,
    help="Audio format",
)
def lossless_check(dir_path: str, format: str, recursive_mode: bool) -> None:
    """Check for lossless audio file"""

    engine = DirectoryAudioLosslessCheckMixin(Path(dir_path).resolve())
    try:
        engine.lossless_check(from_format=format, recursive=recursive_mode)
    except Exception:
        engine.lossless_check_single_thread(from_format=format, recursive=recursive_mode)
    click.echo("Done")


@click.group(name="audio")
def audio_group() -> None:
    """Audio related"""
    pass

audio_group.add_command(lossless_check)
audio_group.add_command(convert_to_mp3)



"""
Absfuyu: Audio
--------------
Audio convert

Version: 5.17.0
Date updated: 20/11/2025 (dd/mm/yyyy)
"""

# Module level
# ---------------------------------------------------------------------------
__all__ = ["DirectoryAudioConvertMixin"]


# Library
# ---------------------------------------------------------------------------
import json
import subprocess
from concurrent.futures import ProcessPoolExecutor, as_completed
from pathlib import Path
from typing import Literal

from absfuyu.core.dummy_func import tqdm
from absfuyu.extra.audio._util import ResultStatus as ConvertStatus
from absfuyu.extra.audio._util import StatusCode
from absfuyu.util import is_command_available
from absfuyu.util.path import DirectoryBase


# Class
# ---------------------------------------------------------------------------
class DirectoryAudioConvertMixin(DirectoryBase):
    """
    Directory - Audio convert to mp3

    - convert_to_mp3
    """

    def __init__(self, source_path, create_if_not_exist=False):
        super().__init__(source_path, create_if_not_exist)
        is_command_available(["ffmpeg"], "ERROR: ffmpeg not installed to PATH")

    def _read_metadata(self, audio_path: Path, /) -> dict:
        """
        Read audio metadata

        Parameters
        ----------
        audio_path : Path
            Path to audio

        Returns
        -------
        dict
            Audio's metadata
        """
        cmd = [
            "ffprobe",
            "-v",
            "quiet",
            "-print_format",
            "json",
            "-show_format",
            str(audio_path.resolve()),
        ]
        result = subprocess.run(cmd, capture_output=True, text=True)
        data = json.loads(result.stdout)

        if "format" in data and "tags" in data["format"]:
            return data["format"]["tags"]
        return {}

    def convert_one(self, audio_path: Path | str, bitrate: Literal["128k", "320k"] = "320k") -> ConvertStatus:
        """
        Convert audio to mp3

        Parameters
        ----------
        audio_path : Path | str
            Path to audio

        bitrate : Literal["128k", "320k"], optional
            Bitrate, by default "320k"

        Returns
        -------
        ConvertStatus
            Result
        """
        audio_path = Path(audio_path)
        mp3_path = audio_path.with_suffix(".mp3")

        if mp3_path.exists():
            return ConvertStatus(StatusCode.SKIP, mp3_path)

        metadata = self._read_metadata(audio_path)

        ffmpeg_cmd = ["ffmpeg", "-y", "-i", str(audio_path.resolve())]

        # Add metadata tags
        for key, value in metadata.items():
            ffmpeg_cmd.extend(["-metadata", f"{key}={value}"])

        ffmpeg_cmd.extend(["-vn", "-codec:a", "libmp3lame", "-b:a", bitrate, str(mp3_path.resolve())])

        subprocess.run(ffmpeg_cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        return ConvertStatus(StatusCode.OK, mp3_path)

    def convert_to_mp3_single_thread(
        self,
        from_format: str = ".flac",
        recursive: bool = True,
        bitrate: Literal["128k", "320k"] = "320k",
    ) -> None:
        """
        Convert audios to .mp3 - Single thread

        Parameters
        ----------
        from_format : str, optional
            Audio format, by default ".flac"

        recursive : bool, optional
            Include audio in child folder, by default True

        bitrate : Literal["128k", "320k"], optional
            Bitrate, by default "320k"
        """
        audios = list(self.source_path.rglob(f"{'**/*'if recursive else '*'}{from_format}"))

        if not audios:
            print(f"No {from_format.upper()} files found.")
            return None

        print(f"Found {len(audios)} {from_format.upper()} files.")

        results: list[ConvertStatus] = []
        for x in tqdm(audios, desc="Converting", unit="file", unit_scale=True):
            results.append(self.convert_one(x, bitrate=bitrate))

        print("\n--- Summary ---")
        for line in results:
            line.print()

    def convert_to_mp3(
        self,
        from_format: str = ".flac",
        workers: int | None = None,
        recursive: bool = True,
        bitrate: Literal["128k", "320k"] = "320k",
    ) -> None:
        """
        Convert audios to .mp3

        Parameters
        ----------
        from_format : str, optional
            Audio format, by default ".flac"

        workers : int | None, optional
            Number of parallel processing threads, by default None

        recursive : bool, optional
            Include audio in child folder, by default True

        bitrate : Literal["128k", "320k"], optional
            Bitrate, by default "320k"
        """
        audios = list(self.source_path.rglob(f"{'**/*'if recursive else '*'}{from_format}"))

        if not audios:
            print(f"No {from_format.upper()} files found.")
            return None

        print(f"Found {len(audios)} {from_format.upper()} files.")

        results: list[ConvertStatus] = []
        with ProcessPoolExecutor(max_workers=workers) as executor:
            futures = {executor.submit(self.convert_one, x, bitrate): x for x in audios}

            for fut in tqdm(
                as_completed(futures),
                total=len(futures),
                desc="Converting",
                unit="file",
                unit_scale=True,
            ):
                results.append(fut.result())

        print("\n--- Summary ---")
        for line in results:
            line.print()



"""
Absfuyu: Tarot
--------------
Tarot stuff


Version: 5.17.0
Date updated: 20/11/2025 (dd/mm/yyyy)

Usage:
------
>>> tarot_deck = Tarot()
>>> print(tarot_deck.random_card())
"""

# Module level
# ---------------------------------------------------------------------------
__all__ = ["Tarot", "TarotCard"]


# Library
# ---------------------------------------------------------------------------
import random

from absfuyu.core import BaseClass, versionadded
from absfuyu.logger import logger
from absfuyu.pkg_data import DataList, DataLoader


# Class
# ---------------------------------------------------------------------------
@versionadded("2.6.0")
class TarotCard:
    """Tarot card"""

    def __init__(
        self,
        name: str,
        rank: int,
        suit: str,
        meanings: dict[str, list[str]],
        keywords: list[str],
        fortune_telling: list[str],
    ) -> None:
        self.name = name.title()
        self.rank = rank
        self.suit = suit
        self.meanings = meanings
        self.keywords = keywords
        self.fortune_telling = fortune_telling

    def __str__(self) -> str:
        return f"{self.__class__.__name__}({self.name})"

    def __repr__(self) -> str:
        return self.__str__()


@versionadded("2.6.0")
class Tarot(BaseClass):
    """Tarot data"""

    def __init__(self) -> None:
        self.data_location = DataList.TAROT

    def __str__(self) -> str:
        return f"{self.__class__.__name__}()"

    @property
    def tarot_deck(self) -> list[TarotCard]:
        """
        Load pickled tarot data

        :rtype: list[TarotCard]
        """
        tarot_data: list = DataLoader(self.data_location).load()
        logger.debug(f"{len(tarot_data)} tarot cards loaded")
        return [
            TarotCard(
                name=x["name"],
                rank=x["rank"],
                suit=x["suit"],
                meanings=x["meanings"],
                keywords=x["keywords"],
                fortune_telling=x["fortune_telling"],
            )
            for x in tarot_data
        ]

    def random_card(self) -> TarotCard:
        """
        Pick a random tarot card

        Returns
        -------
        TarotCard
            Random Tarot card
        """
        return random.choice(self.tarot_deck)
